"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

CliResponse module. Contains CliResponse class, which is the object returned by the command API.
"""


import logging

import icetea_lib.LogManager as LogManager
from icetea_lib.Searcher import verify_message
from icetea_lib.TestStepError import TestStepFail


# Disable "value x is unsubscriptable"
# pylint: disable=E1136


# Command Response
class CliResponse(object):
    """
    CliResponse class. Object returned by the command api when a command has finished.
    """

    # pylint:disable=invalid-name
    # Constructor
    def __init__(self):
        try:
            self.logger = LogManager.get_bench_logger()
        except KeyError:
            self.logger = logging.getLogger("bench")
        self.timeout = False
        self.timedelta = 0
        self.retcode = None
        self.lines = []
        self.traces = []
        self.parsed = None

    def __str__(self):
        return str(self.retcode)

    def success(self):
        """
        Check if this is a response to a successful command
        :return: True or False
        """
        return self.retcode == 0

    def fail(self):
        """
        Check if this is a response to a failed command
        :return: True or False
        """
        return self.retcode != 0

    # verify that response message is expected
    def verify_message(self, expected_response, break_in_fail=True):
        """
        Verifies that expected_response is found in self.lines.

        :param expected_response: response or responses to look for. Must be list or str.
        :param break_in_fail: If set to True,
        re-raises exceptions caught or if message was not found
        :return: True or False
        :raises: LookupError if message was not found and break_in_fail was True. Other exceptions
        might also be raised through searcher.verify_message.
        """
        ok = True
        try:
            ok = verify_message(self.lines, expected_response)
        except (TypeError, LookupError) as inst:
            ok = False
            if break_in_fail:
                raise inst
        if ok is False and break_in_fail:
            raise LookupError("Unexpected message found")
        return ok

    def verify_trace(self, expected_traces, break_in_fail=True):
        """
        Verifies that expectedResponse is found in self.traces

        :param expected_traces: response or responses to look for. Must be list or str.
        :param break_in_fail: If set to True, re-raises exceptions caught or if message was
        not found
        :return: True or False
        :raises: LookupError if message was not found and breakInFail was True. Other Exceptions
        might also be raised through searcher.verify_message.
        """
        ok = True
        try:
            ok = verify_message(self.traces, expected_traces)
        except (TypeError, LookupError) as inst:
            ok = False
            if break_in_fail:
                raise inst
        if ok is False and break_in_fail:
            raise LookupError("Unexpected message found")
        return ok

    def set_response_time(self, seconds):
        """
        Set response time in seconds.

        :param seconds: integer
        :return: Nothing
        """
        self.timedelta = seconds

    def verify_response_duration(self, expected=None, zero=0, threshold_percent=0,
                                 break_in_fail=True):
        """
        Verify that response duration is in bounds.

        :param expected: seconds what is expected duration
        :param zero: seconds if one to normalize duration before calculating error rate
        :param threshold_percent: allowed error in percents
        :param break_in_fail: boolean, True if raise TestStepFail when out of bounds
        :return: (duration, expected duration, error)
        """
        was = self.timedelta - zero
        error = abs(was/expected)*100.0 - 100.0 if expected > 0 else 0
        msg = "should: %.3f, was: %.3f, error: %.3f %%" % (expected, was, error)
        self.logger.debug(msg)
        if abs(error) > threshold_percent:
            msg = "Thread::wait error(%.2f %%) was out of bounds (%.2f %%)" \
                  % (error, threshold_percent)
            self.logger.debug(msg)
            if break_in_fail:
                raise TestStepFail(msg)
        return was, expected, error

    def verify_response_time(self, expected_below):
        """
        Verify that response time (time span between request-response) is reasonable.

        :param expected_below: integer
        :return: Nothing
        :raises: ValueError if timedelta > expected time
        """
        if self.timedelta > expected_below:
            raise ValueError("Response time is more (%f) than expected (%f)!"
                             % (self.timedelta, expected_below))



# pylint: disable=missing-docstring

"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import unittest

import mock

from icetea_lib.AllocationContext import AllocationContext, AllocationContextList
from icetea_lib.DeviceConnectors.Dut import DutConnectionError
from icetea_lib.ResourceProvider.exceptions import ResourceInitError


class AllocContextTestcase(unittest.TestCase):

    def test_set(self):
        con1 = AllocationContext("id1", "al_id1", {"data": "data1"})
        con1["test"] = "test"
        self.assertEqual(con1.get_alloc_data()["test"], "test")
        con1.set("test", "test2")
        self.assertEqual(con1.get_alloc_data()["test"], "test2")

    def test_get(self):
        con1 = AllocationContext("id1", "al_id1", {"data": "data1"})
        con1.set("test", "test3")
        self.assertEqual(con1.get("test"), "test3")
        self.assertEqual(con1["test"], "test3")


class AllocContextListTestcase(unittest.TestCase):

    def setUp(self):
        self.nulllogger = logging.getLogger("test")
        self.nulllogger.addHandler(logging.NullHandler())

    def test_append_and_len(self):
        con1 = AllocationContext("id1", "al_id1", {"data": "data1"})
        con2 = AllocationContext("id2", "al_id2", {"data": "data2"})
        con_list = AllocationContextList(self.nulllogger)
        self.assertEqual(len(con_list), 0)
        con_list.append(con1)
        self.assertEqual(len(con_list), 1)
        con_list.append(con2)
        self.assertEqual(len(con_list), 2)

    def test_get_and_set_item(self):
        con1 = AllocationContext("id1", "al_id1", {"data": "data1"})
        con2 = AllocationContext("id2", "al_id2", {"data": "data2"})
        con_list = AllocationContextList(self.nulllogger)
        con_list.append(con1)
        con_list.append(con2)

        self.assertEqual(con_list[0].resource_id, "id1")
        self.assertEqual(con_list[1].alloc_id, "al_id2")
        con_list[0] = con2
        self.assertEqual(con_list[0].resource_id, "id2")

        with self.assertRaises(IndexError):
            er = con_list[2]  # pylint: disable=unused-variable,invalid-name
        with self.assertRaises(IndexError):
            er = con_list[-1]  # pylint: disable=invalid-name
        with self.assertRaises(IndexError):
            con_list[-1] = "test"
        with self.assertRaises(IndexError):
            con_list[3] = "test"

        with self.assertRaises(TypeError):
            er = con_list["test"]  # pylint: disable=unused-variable,invalid-name
        with self.assertRaises(TypeError):
            con_list["test"] = "test"

    def test_open_dut_connections(self):
        con_list = AllocationContextList(self.nulllogger)
        # Setup mocked duts
        dut1 = mock.MagicMock()
        dut1.start_dut_thread = mock.MagicMock()
        dut1.start_dut_thread.return_value = "ok"
        dut1.open_dut = mock.MagicMock()
        dut1.open_dut.return_value = "ok"
        dut1.close_dut = mock.MagicMock()
        dut1.close_connection = mock.MagicMock()
        dut2 = mock.MagicMock()
        dut2.start_dut_thread = mock.MagicMock()
        dut2.start_dut_thread.side_effect = [DutConnectionError]
        dut2.open_dut = mock.MagicMock()
        dut2.open_dut.return_value = "ok"
        dut2.close_dut = mock.MagicMock()
        dut2.close_connection = mock.MagicMock()

        con_list.duts = [dut1]
        con_list.open_dut_connections()
        dut1.start_dut_thread.assert_called()
        dut1.open_dut.assert_called()

        con_list.duts.append(dut2)
        with self.assertRaises(DutConnectionError):
            con_list.open_dut_connections()
        dut2.start_dut_thread.assert_called()
        dut2.close_dut.assert_called()
        dut2.close_connection.assert_called()

    @mock.patch("icetea_lib.AllocationContext.os.path.isfile")
    @mock.patch("icetea_lib.AllocationContext.AllocationContextList.get_build")
    def test_check_flashing_need(self, mock_get_build, mock_isfile):
        con_list = AllocationContextList(self.nulllogger)
        mock_get_build.return_value = "test_name.bin"
        mock_isfile.return_value = True
        self.assertTrue(con_list.check_flashing_need("hardware", "test_build", False))
        mock_get_build.return_value = "test_name.hex"
        self.assertTrue(con_list.check_flashing_need("hardware", "test_build", False))
        mock_get_build.return_value = "test_name"
        self.assertFalse(con_list.check_flashing_need("hardware", "test_build", False))

        self.assertTrue(con_list.check_flashing_need("hardware", "test_build", True))

        mock_isfile.return_value = False
        with self.assertRaises(ResourceInitError):
            con_list.check_flashing_need("hardware", "test_build", False)


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

LogManager module contains all logging related methods,
classes and helpers used by Icetea.
"""

import os
import time
import logging
import logging.config
import re
import six

COLORS = True
try:
    import coloredlogs
except ImportError:
    coloredlogs = None
    COLORS = False

# Heavy use of global statement in this module for good reasons. Disable global-statement warning.
# pylint: disable=global-statement

LOGPATHDIR = None  # Path to run log directory
LOGTCDIR = None
LOGTCNAME = None
LOGGERS = {}
LOGFILES = []
GLOBAL_LOGFILES = []
REPEATNUM = 0
STANDALONE_LOGGING = True
VERBOSE_ON = True
SILENT_ON = False
COLOR_ON = False
TRUNCATE_LOG = True

LEVEL_FORMATS = dict(
    debug=dict(color='white'),
    info=dict(color='green'),
    verbose=dict(color='blue'),
    warning=dict(color='yellow'),
    error=dict(color='red'),
    critical=dict(color='red'))

FIELD_STYLES = dict(
    asctime=dict(color='cyan'),
    hostname=dict(color='magenta'),
    levelname=dict(color='black'),
    programname=dict(color='white'),
    name=dict(color='blue'),
    threadName=dict(color='blue'),
    source=dict(color='blue'))


class BenchLoggerAdapter(logging.LoggerAdapter):
    """
    Adapter to add field 'extra' to logger.
    """
    def process(self, msg, kwargs):
        if not "extra" in kwargs:
            kwargs["extra"] = {}
        kwargs["extra"]["source"] = self.extra["source"]
        return msg, kwargs


class BenchFormatterWithType(object):  # pylint: disable=too-few-public-methods
    """
    Bench logger formatter.
    """
    def __init__(self, color=False):
        if not color:
            self._formatter = logging.Formatter(
                "%(asctime)s.%(msecs)03d | %(source)s %(type)s %(threadName)s: "
                "%(message)s", "%H:%M:%S")
        else:
            self._formatter = coloredlogs.ColoredFormatter(
                "%(asctime)s.%(msecs)03d | %(source)s %(type)s %(threadName)s: "
                "%(message)s", "%H:%M:%S", LEVEL_FORMATS, FIELD_STYLES)

    def format(self, record):
        """
        Format record with formatter.
        :param record: Record to format
        :return: Formatted record
        """
        if not hasattr(record, "type"):
            record.type = "   "
        return self._formatter.format(record)


def remove_handlers(logger):
    # TODO: Issue related to placeholder logger objects appearing in some rare cases. Check below
    # required as a workaround
    """
    remove handlers from logger.

    :param logger: Logger whose handlers to remove
    :return:
    """
    if hasattr(logger, "handlers"):
        for handler in logger.handlers[::-1]:
            try:
                if isinstance(handler, logging.FileHandler):
                    handler.close()
                logger.removeHandler(handler)
            except:  # pylint: disable=bare-except
                import traceback
                traceback.print_exc()
                break


def get_base_dir():
    """
    Return the directory where logs for this run will be stored
    """
    return LOGPATHDIR


def get_testcase_log_dir():
    """
    Return the directory where current testcase's logs will be stored
    """
    return LOGTCDIR


def get_testcase_logfilename(logname, prepend_tc_name=False):
    """
    Return filename for a logfile, filename will contain the actual path + filename.

    :param logname: Name of the log including the extension, should describe
    what it contains (eg. "device_serial_port.log")
    :param prepend_tc_name: Boolean, if True, prepends the filename with the
    testcase name
    """
    logdir = get_testcase_log_dir()
    if prepend_tc_name:
        global LOGTCNAME
        logname = LOGTCNAME + "_" + logname
    fname = os.path.join(logdir, logname)
    LOGFILES.append(fname)
    return fname


def get_base_logfilename(logname):
    """ Return filename for a logfile, filename will contain the actual path + filename

    :param logname: Name of the log including the extension, should describe
    what it contains (eg. "device_serial_port.log")
    """
    logdir = get_base_dir()
    fname = os.path.join(logdir, logname)
    GLOBAL_LOGFILES.append(fname)
    return fname


def get_logger(name):
    """
    Return a logger instance for given name
    """
    return logging.getLogger(name)


def get_file_logger(name, formatter=None):
    """
    Return a file logger that will log into a file located in the testcase log
    directory. Anything logged with a file logger won't be visible in the
    console or any other logger.

    :param name, Name of the logger, eg. the module name
    """
    if name is None or name == "":
        raise ValueError("Can't make a logger without name")

    logger = logging.getLogger(name)
    logger.propagate = False
    remove_handlers(logger)
    logger.setLevel(logging.INFO)

    if formatter is None:
        formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", "%H:%M:%S")
    func = get_testcase_logfilename(name + ".log")
    handler = _get_filehandler_with_formatter(func, formatter)
    logger.addHandler(handler)

    return logger


def get_resourceprovider_logger(name=None, short_name=" ", log_to_file=True):
    """
    Get logger for ResourceProvider and related classes.

    :param name: Name of the logger
    :param short_name: Shorthand name
    :param log_to_file: If True enable logging to file.
    :return: logging.Logger
    """

    global LOGGERS
    loggername = name
    # Return existing logger if one exists
    if loggername in LOGGERS:
        # Check if short_name matches the existing one, if not update it
        if isinstance(LOGGERS[loggername], BenchLoggerAdapter):
            if ("source" not in LOGGERS[loggername].extra or
                    LOGGERS[loggername].extra["source"] != short_name):
                LOGGERS[loggername].extra["source"] = short_name
        return LOGGERS[loggername]

    logger = logging.getLogger(loggername)
    logger.propagate = False
    remove_handlers(logger)
    logger.setLevel(logging.DEBUG)
    if TRUNCATE_LOG:
        logger.addFilter(ContextFilter())

    # Filehandler for logger
    if log_to_file:
        func = get_base_logfilename(name + ".log")
        handler = _get_filehandler_with_formatter(func, BenchFormatterWithType())
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    cbh = logging.StreamHandler()
    cbh.formatter = BenchFormatterWithType(COLOR_ON)
    if VERBOSE_ON:
        cbh.setLevel(logging.DEBUG)
    elif SILENT_ON:
        cbh.setLevel(logging.WARN)
    else:
        cbh.setLevel(logging.INFO)
    logger.addHandler(cbh)

    LOGGERS[loggername] = BenchLoggerAdapter(logger, {"source": short_name})

    return LOGGERS[loggername]


def get_bench_logger(name=None, short_name="  ", log_to_file=True):
    """
    Return a logger instance for given name. The logger will be a child
    of the bench logger, so anything that is logged to it, will be also logged
    to bench logger. If a logger with the given name doesn't already exist,
    create it using the given parameters.

    :param name: Name of the logger
    :param short_name: A short name (preferably 3 characters) describing the logger
    :param log_to_file: Boolean, if True the logger will also log to a file "name.log"
    """
    global LOGGERS

    # Get the root bench logger if name is none or empty or bench
    if name is None or name == "" or name == "bench":
        return LOGGERS["bench"]

    loggername = "bench." + name
    # Return existing logger if one exists
    if loggername in LOGGERS:
        # Check if short_name matches the existing one, if not update it
        if isinstance(LOGGERS[loggername], BenchLoggerAdapter):
            if ("source" not in LOGGERS[loggername].extra or
                    LOGGERS[loggername].extra["source"] != short_name):
                LOGGERS[loggername].extra["source"] = short_name
        return LOGGERS[loggername]

    logger = logging.getLogger(loggername)
    logger.propagate = True
    remove_handlers(logger)
    logger.setLevel(logging.DEBUG)
    if TRUNCATE_LOG:
        logger.addFilter(ContextFilter())

    # Filehandler for logger
    if log_to_file:
        func = get_testcase_logfilename(name + ".log")
        handler = _get_filehandler_with_formatter(func, BenchFormatterWithType())
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    LOGGERS[loggername] = BenchLoggerAdapter(logger, {"source": short_name})

    return LOGGERS[loggername]


def get_dummy_logger():
    """
    Get dummy logger.

    :return: logger with NullHandler
    """
    logger = logging.getLogger("dummy")
    logger.propagate = False
    logger.addHandler(logging.NullHandler())
    return logger


def get_logfiles():
    """
    Return a list of logfiles with relative paths from the log
    root directory
    """
    logfiles = [f for f in LOGFILES]
    logfiles.extend(GLOBAL_LOGFILES)
    return logfiles


def set_level(name, level):
    """
    Set level for given logger.

    :param name: Name of logger to set the level for
    :param level: The new level, see possible levels from python logging library
    """
    if name is None or name == "" or name == "bench":
        logging.getLogger("bench").setLevel(level)
    loggername = "bench." + name
    logging.getLogger(loggername).setLevel(level)


# pylint: disable=too-many-arguments
def init_base_logging(directory="./log", verbose=False, silent=False, color=False, no_file=False,
                      truncate=True):
    """
    Initialize the Icetea logging by creating a directory to store logs
    for this run and initialize the console logger for Icetea itself.

    :param directory: Directory where to store the resulting logs
    :param verbose: Log level debug
    :param silent: Log level warning
    :param no_file: Log to file
    :param color: Log coloring
    :param truncate: Log truncating
    :return: Nothing
    """
    global LOGPATHDIR
    global STANDALONE_LOGGING
    global TRUNCATE_LOG
    global COLOR_ON
    global SILENT_ON
    global VERBOSE_ON

    LOGPATHDIR = os.path.join(directory, time.strftime("%Y-%m-%d_%H%M%S"))
    if not os.path.exists(LOGPATHDIR) and not no_file:
        os.makedirs(LOGPATHDIR)

    # Initialize the simple console logger for IceteaManager
    icetealogger = logging.getLogger("icetea")
    icetealogger.propagate = False
    icetealogger.setLevel(logging.DEBUG)
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", "%H:%M:%S")
    if not color:
        stream_handler.setFormatter(formatter)
    elif color and not COLORS:
        raise ImportError("Missing coloredlogs module. Please install with"
                          "pip to use colors in logging.")
    else:
        stream_handler.setFormatter(coloredlogs.ColoredFormatter(
            "%(asctime)s.%(msecs)03d %(message)s", "%H:%M:%S"))

    if color and COLORS:
        COLOR_ON = color
    SILENT_ON = silent
    VERBOSE_ON = verbose
    if not no_file:
        file_handler = logging.FileHandler(get_base_logfilename("icetea.log"))
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        icetealogger.addHandler(file_handler)
    if verbose:
        stream_handler.setLevel(logging.DEBUG)
    elif silent:
        stream_handler.setLevel(logging.WARN)
    else:
        stream_handler.setLevel(logging.INFO)
    icetealogger.addHandler(stream_handler)
    TRUNCATE_LOG = truncate
    if TRUNCATE_LOG:
        icetealogger.addFilter(ContextFilter())

    STANDALONE_LOGGING = False


def init_testcase_logging(name, verbose=True, silent=False, color=False,
                          truncate=True):
    """
    Initialize testcase logging and default loggers. First removes any
    existing testcase loggers, creates a new directory for testcase logs under
    the root logging directory, initializes the default bench logger and
    removes any child loggers it may have.

    :param name: Name of the testcase
    :param verbose: Log level debug
    :param silent: Log level warning
    :param color: Log coloring
    :param truncate: Log truncating
    """
    global LOGTCDIR
    global LOGTCNAME
    global LOGGERS
    global REPEATNUM
    global VERBOSE_ON
    global SILENT_ON
    global TRUNCATE_LOG

    if name is None or not isinstance(name, str) or not name:
        raise EnvironmentError("Invalid testcase name for logging configuration")

    if STANDALONE_LOGGING:
        init_base_logging()

    finish_testcase_logging()

    # Generate filename friendly base name for logs from testcase name
    LOGTCDIR = re.sub(r'[^\w\-_\. ]', '_', name)
    LOGTCNAME = LOGTCDIR
    tempname = os.path.join(LOGPATHDIR, LOGTCDIR + "_0")

    # Check if tc folder already exists. If it exists, we look for a new folder
    # by appending integers to the folder name
    while os.path.exists(tempname):
        tempname = os.path.join(LOGPATHDIR, LOGTCDIR + "_{}".format(REPEATNUM))
        REPEATNUM = REPEATNUM + 1
    LOGTCDIR = tempname
    REPEATNUM = 0

    if not os.path.exists(LOGTCDIR):
        os.makedirs(LOGTCDIR)

    # Get the bench logger and remove all existing handlers
    benchlogger = logging.getLogger("bench")
    benchlogger.propagate = False
    remove_handlers(benchlogger)
    benchlogger.setLevel(logging.DEBUG)

    TRUNCATE_LOG = truncate
    if TRUNCATE_LOG:
        benchlogger.addFilter(ContextFilter())

    # Console handler for bench
    stream_handler = logging.StreamHandler()
    if color and not COLORS:
        raise ImportError("Missing module: coloredlogs. Please install"
                          "with pip to use colors in logging.")
    stream_handler.formatter = BenchFormatterWithType(color)
    VERBOSE_ON = verbose
    SILENT_ON = silent
    if verbose:
        stream_handler.setLevel(logging.DEBUG)
    elif silent:
        stream_handler.setLevel(logging.WARN)
    else:
        stream_handler.setLevel(logging.INFO)
    benchlogger.addHandler(stream_handler)
    # file handler for bench and all child loggers
    func = get_testcase_logfilename("bench.log")
    handler = _get_filehandler_with_formatter(func, BenchFormatterWithType())
    handler.setLevel(logging.DEBUG)
    benchlogger.addHandler(handler)
    LOGGERS["bench"] = BenchLoggerAdapter(benchlogger, {"source": "TC"})


def finish_testcase_logging():
    """
    Finalize testcase logging by removing loggers
    """
    del LOGFILES[:]
    if LOGGERS:
        # From local list of loggers, retrieve one(first one)
        vlst = list(LOGGERS.values())
        logr = vlst[0].logger
        # Retrieve logging module manager from logger
        manager = logr.manager
        # Manager holds dict with hard references to all logger instances
        # produced in this python session
        for name, logger in list(manager.loggerDict.items()):
            # Only bench loggers are required to be non-persistent between
            # testcases
            if 'bench' in name:
                # Remove handlers from logger
                remove_handlers(logger)
                # Delete logger instance
                del manager.loggerDict[name]

    # Clear local logger dictionary
    LOGGERS.clear()


def _get_filehandler_with_formatter(logname, formatter=None):
    """
    Return a logging FileHandler for given logname using a given
    logging formatter.

    :param logname: Name of the file where logs will be stored, ".log"
    extension will be added
    :param formatter: An instance of logging.Formatter or None if the default
    should be used
    :return: FileHandler
    """
    handler = logging.FileHandler(logname)
    if formatter is not None:
        handler.setFormatter(formatter)
    return handler


class ContextFilter(logging.Filter):  # pylint: disable=too-few-public-methods
    """
    Filter for filtering logging messages and truncating them after MAXIMUM_LENGTH has been reached.
    """
    MAXIMUM_LENGTH = 10000
    REVEAL_LENGTH = 50

    def filter(self, record):
        """
        Filter record
        :param record: Record to filter
        :return:
        """
        def modify(value):
            """
            Modify logged record, truncating it to max length and logging remaining length
            :param value: Record to modify
            :return:
            """
            if isinstance(value, six.string_types):
                if len(value) < ContextFilter.MAXIMUM_LENGTH:
                    return value

                try:
                    return "{}...[{} more bytes]".format(
                        value[:ContextFilter.REVEAL_LENGTH],
                        len(value) - ContextFilter.REVEAL_LENGTH)
                except UnicodeError:
                    return "{}...[{} more bytes]".format(
                        repr(value[:ContextFilter.REVEAL_LENGTH]),
                        len(value) - ContextFilter.REVEAL_LENGTH)
            elif isinstance(value, six.binary_type):
                return "{}...[{} more bytes]".format(
                    repr(value[:ContextFilter.REVEAL_LENGTH]),
                    len(value) - ContextFilter.REVEAL_LENGTH)

            else:
                return value

        record.msg = traverse_json_obj(record.msg, callback=modify)
        return True


def traverse_json_obj(obj, path=None, callback=None):
    """
    Recursively loop through object and perform the function defined
    in callback for every element. Only JSON data types are supported.

    :param obj: object to traverse
    :param path: current path
    :param callback: callback executed on every element
    :return: potentially altered object
    """
    if path is None:
        path = []

    if isinstance(obj, dict):
        value = {k: traverse_json_obj(v, path + [k], callback)
                 for k, v in obj.items()}
    elif isinstance(obj, list):
        value = [traverse_json_obj(elem, path + [[]], callback)
                 for elem in obj]
    else:
        value = obj

    if callback is None:
        return value
    return callback(value)



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


ReportHtml module, contains ReportHtml class that implement html reporting.
"""

from datetime import datetime
import os

from yattag import Doc

from icetea_lib.tools.tools import hex_escape_str, get_fw_version, get_fw_name
from icetea_lib.Reports.ReportBase import ReportBase


class ReportHtml(ReportBase):
    """
    ReportHtml class. Uses yattag to generate html reports of results.
    """
    def __init__(self, results):
        ReportBase.__init__(self, results)

    def generate(self, *args, **kwargs):
        """
        Implementation for the generate method defined in ReportBase.
        Generates a html report and saves it.

        :param args: 1 argument, which is the filename
        :param kwargs: 3 keyword arguments with keys 'title', 'heads' and 'refresh'
        :return: Nothing.
        """
        title = kwargs.get("title")
        heads = kwargs.get("heads")
        refresh = kwargs.get("refresh")
        filename = args[0]
        report = self._create(title, heads, refresh, path_start=os.path.dirname(filename))
        ReportHtml.save(report, filename)

    # pylint: disable=too-many-statements
    def _create(self, title, heads, refresh=None, path_start=None):
        """
        Internal create method, uses yattag to generate a html document with result data.

        :param title: Title of report
        :param heads: Headers for report
        :param refresh: If set to True, adds a HTTP-EQUIV="refresh" to the report
        :param path_start: path to file where this is report is to be stored.
        :return: yattag document.
        """
        # TODO: Refactor to make less complex
        doc, tag, text = Doc().tagtext()
        doc.asis('<!DOCTYPE html>')
        heads["Date"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        heads["Pass rate"] = self.results.pass_rate()

        with tag('html'):
            with tag('head'):
                doc.asis(self.head)
                if refresh:
                    doc.asis('<META HTTP-EQUIV="refresh" CONTENT="' + str(refresh) + '">')
            with tag('body', id='body'):
                with tag('h1'):
                    text(title)
                with tag('table'):
                    for head in heads:
                        with tag('tr'):
                            with tag('th', width="100px"):
                                text(head)
                            with tag('td'):
                                text(heads[head])
                    with tag('tr'):
                        with tag('th'):
                            text('Executed')
                        with tag('td'):
                            text(str(self.summary["count"]))
                    with tag('tr'):
                        with tag('th'):
                            text('Pass:')
                        with tag('td'):
                            text(str(self.summary["pass"]))
                    with tag('tr'):
                        with tag('th'):
                            text('Fails:')
                        with tag('td'):
                            text(str(self.summary["fail"]))
                    with tag('tr'):
                        with tag('th'):
                            text('inconclusive:')
                        with tag('td'):
                            text(str(self.summary["inconclusive"]))
                    with tag('tr'):
                        with tag('th'):
                            text('Skip:')
                        with tag('td'):
                            text(str(self.summary["skip"]))
                    with tag('tr'):
                        with tag('th'):
                            text('Duration:')
                        with tag('td'):
                            text(self.duration_to_string(self.summary["duration"]))
                    with tag('tr'):
                        with tag('th'):
                            text('{} version:'.format(get_fw_name()))
                        with tag('td'):
                            text(get_fw_version())

                with tag('table', style='border-collapse: collapse;'):
                    with tag('tr'):
                        with tag('th'):
                            text("Test Case")
                        with tag('th'):
                            text("Verdict")
                        with tag('th'):
                            text("Fail Reason")
                        with tag('th'):
                            text("Skip Reason")
                        with tag('th'):
                            text("Duration")
                    for result in self.results:
                        if result.success:
                            klass = 'item_pass'
                        elif result.inconclusive:
                            klass = 'item_inconc'
                        else:
                            klass = 'item_fail'
                        with tag('tr', klass='item %s' % klass, onclick='showhide(this)'):
                            with tag('td', width="200px"):
                                text(result.get_tc_name())
                            with tag('td', width="100px"):
                                if result.success:
                                    color = 'green'
                                elif result.failure:
                                    color = 'red'
                                else:
                                    color = 'black'
                                with tag('font', color=color):
                                    text(result.get_verdict())
                            with tag('td', width="350px"):
                                text(hex_escape_str(result.fail_reason))
                            with tag('td', width="300px"):
                                text(result.skip_reason if result.skipped() else "")
                            with tag('td', width="100px"):
                                text(str(result.duration))
                        with tag('tr', klass='info hidden'):
                            with tag('td', colspan="5"):
                                if hasattr(result, 'tc_git_info') and \
                                        result.tc_git_info and \
                                        "scm_link" in result.tc_git_info:
                                    # add tc git info only when available
                                    link = result.tc_git_info['scm_link']
                                    with tag('a', href=link):
                                        text(link)
                                    doc.stag('br')
                                for fil in result.logfiles:
                                    filepath = os.path.relpath(fil, path_start)
                                    with tag('a', href=filepath):
                                        text(filepath)
                                    doc.stag('br')

        return doc.getvalue()

    @staticmethod
    def save(htmldoc, filename):
        """
        Static method which saves htmldoc with filename filename.

        :param htmldoc: yattag Document
        :param filename: file name/path
        :return: Nothing
        """
        with open(filename, 'w') as fil:
            fil.write(htmldoc)

    head = """
    <style type="text/css">
        .item {
            border-top: 2px solid #505050;
            border-bottom: 1px solid #A0A0A0;
            border-left: 1px solid #A0A0A0;
            border-right: 1px solid #A0A0A0;
        }

        .item:hover {
            cursor: pointer;
        }

        .item_fail {
            background-color: #FFC0C0;
        }

        .item_inconc {
            background-color: #FFD733;
        }

        .item_inconc:hover {
            background-color: #FCFF58;
        }

        .item_fail:hover {
            background-color: #FFCFCF;
        }

        .item_pass {
            background-color: #C0FFC0;
        }

        .item_pass:hover {
            background-color: #CFFFCF;
        }

        .info {
            display: ;
            background: transparent;
            border-left: 1px solid #A0A0A0;
            border-right: 1px solid #A0A0A0;
            border-bottom: 2px solid #505050;
        }

        .hidden {
            display: None;
        }

        .visible_info td {
            padding-left: 1em;
        }

    </style>
    <script type="text/javascript">
        function showhide(which) {
            n = which.parentNode.rows[which.rowIndex + 1];
            if (n != null) {
                if (n.classList.contains("hidden")) {
                    n.classList.remove("hidden")
                }
                else {
                    n.classList.add("hidden")
                }
            }
        }
    </script>"""


# pylint: disable=missing-docstring
# -*- coding: utf-8 -*-

"""
Copyright 2019 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import subprocess
import unittest
import platform
import os

from icetea_lib.IceteaManager import ExitCodes


class GenericProcessTestcase(unittest.TestCase):
    @unittest.skipIf(
        platform.system() == "Windows",
        "Launching process in Windows ends in icetea warning"
        "\"This OS is not supporting select.poll() or select.kqueue()\"")
    def test_quick_process(self):
        # Run testcase
        icetea_cmd = ["python", "icetea.py",
                      "--clean",
                      "--silent",
                      "--failure_return_value",
                      "--tc", "test_quick_process",
                      "--tcdir", '"{}"'.format(os.path.join("test", "tests"))]

        # Shouldn't need to join the argument array,
        # but it doesn't work for some reason without it (Python 2.7.12).
        retcode = subprocess.call(args=" ".join(icetea_cmd), shell=True)

        # Check success
        self.assertEqual(
            retcode, ExitCodes.EXIT_SUCCESS,
            "Non-success returncode {} returned.".format(retcode))

    def tearDown(self):
        # Run with --clean to clean up
        subprocess.call(
            "python icetea.py --clean ",
            shell=True)


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import traceback
import sys

from icetea_lib.LogManager import get_bench_logger


class TestStepError(Exception):
    '''
    TestStepError exception is used in case where something
    very fatal unexpected happens in test environment or serial port connection dies.
    '''
    def __init__(self, message="TestStepError"):
        super(TestStepError, self).__init__(message)
        self.logger = None

    # print detailed info
    def detailed_info(self):
        """
        Print detailed exception information from traceback
        """
        _, _, trace = sys.exc_info()
        filename, linenumber, functionname, line = traceback.extract_tb(trace)[-2]
        self.logger = get_bench_logger()
        self.logger.error("Exception details: ")
        self.logger.error("TC Name: " + str(filename))
        self.logger.error("Function name: " + str(functionname))
        self.logger.error("Line number: " + str(linenumber))
        self.logger.error("Line: " + str(line))


class InconclusiveError(Exception):
    """
    Inconclusive error exception, used for showing errors such as
    broken environments and missing requirements.
    """
    pass


class TestStepFail(Exception):
    '''
    TestStepFail exception is used when failure causes because of
    device/software under test, and probably not related to test environment
    '''
    pass


class TestStepTimeout(TestStepFail):
    """
    TestStepTimeout, failure for  showing timeouts.
    """
    pass



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

ResourceProvider module. Handling of allocations and resource configurations is done here.
"""

import json

import six

import icetea_lib.LogManager as LogManager
from icetea_lib.ResourceProvider.exceptions import ResourceInitError
from icetea_lib.ResourceProvider.Allocators.exceptions import AllocationError
from icetea_lib.tools.tools import Singleton


@six.add_metaclass(Singleton)
class ResourceProvider(object):
    """
    Singleton ResourceProvider class. ResourceProvider is common for the entire run and it
    handles allocation of resources based on requested resource configurations. It also
    determines which allocator is used for allocating these resources.
    """

    __metaclass__ = Singleton

    def __init__(self, args):
        # @TODO: Refactor args into some separate configuration class maybe?
        self.args = args
        self.allocator = None
        self.jsonconf = None
        no_file = False if self.args.list or self.args.listsuites else True
        self.logger = LogManager.get_resourceprovider_logger("ResourceProvider", "RSP", no_file)
        self._pluginmanager = None

    def set_pluginmanager(self, pluginmanager):
        self._pluginmanager = pluginmanager

    def __del__(self):
        if self.allocator:
            self.allocator.cleanup()
            self.allocator = None

    def resolve_configuration(self, conf, resource_configuration):
        """
        Resolve the configuration from given JSON encoded configuration data.

        :param conf: JSON encoded configuration
        :param resource_configuration: ResourceConfig object
        """
        if not self.logger:
            self.logger = LogManager.get_resourceprovider_logger("ResourceProvider", "RSP")
        self.jsonconf = conf
        resource_configuration.resolve_configuration(conf)

    def allocate_duts(self, resource_configuration):
        """
        Initialize DUT's.

        :param resource_configuration: ResourceConfig
        :return: List of DUT objects
        :raises: ResourceInitError
        """
        cnt_hrdwr = resource_configuration.count_hardware()
        if not self.allocator:
            self.allocator = self.__get_allocator()
        try:
            # Try to allocate resources and instantiate DUT's
            dut_conf_list = resource_configuration.get_dut_configuration()
            if dut_conf_list:
                self.logger.debug("Allocating duts with the following configurations:")
                for conf in dut_conf_list:
                    self.logger.debug(conf.get_requirements())
            self.logger.info("Allocating {} duts.".format(resource_configuration.count_duts()))
            allocations = self.allocator.allocate(resource_configuration, args=self.args)
            self.logger.info("Allocation successful")
        except AllocationError as error:
            raise ResourceInitError(error)

        return allocations

    def cleanup(self):
        """
        Clean up allocator at the end of the run.

        :return: Nothing
        """
        self.logger.debug("Cleaning up ResourceProvider.")
        if self.allocator:
            self.logger.debug("Cleaning up allocator.")
            self.allocator.cleanup()

    def __get_allocator(self):
        """
        Internal method for determining which allocator is needed for this run.

        :return: BaseAllocator
        :raises: ResourceInitError
        """
        allocator_name = self.args.allocator
        allocator_cfg_file = self.args.allocator_cfg
        allocator_cfg = dict()
        if allocator_cfg_file:
            allocator_cfg = self._read_allocator_config(allocator_name, allocator_cfg_file)
        allocator = self._pluginmanager.get_allocator(allocator_name)
        if allocator is None:
            raise ResourceInitError("Unable to load allocator {}".format(allocator_name))
        self.logger.debug("Using allocator %s", allocator_name)
        return allocator(self.args, None, allocator_cfg)

    def get_my_duts(self):
        """
        Get my duts.

        :return: list of duts
        """
        # TODO: Is this function still used somewhere? There are no usages in this project at least.
        if self.args.my_duts:
            myduts = self.args.my_duts.split(',')
            return myduts
        return None

    def _read_allocator_config(self, allocator_name, allocator_cfg_file):
        """
        Read configuration for allocator from a json file.
        Json file needs to have an object that contains key allocator_name (if you want to use
        the same config for each allocator for example)

        :param allocator_name: Name of the allocator.
        :param allocator_cfg_file: absolute path to the json config file to use.
        :return: dict
        :raises: ResourceInitError if config file not found.
        """
        allocator_config = dict()
        self.logger.debug("Reading allocator configuration from {}".format(allocator_cfg_file))
        try:
            with open(allocator_cfg_file, "r") as cfg_file:
                try:
                    data = json.load(cfg_file)
                except ValueError as e:
                    self.logger.error(e)
                    raise ResourceInitError("Failed to decode json "
                                            "from allocator config file {}".format(cfg_file))

                if allocator_name in data:
                    allocator_config = data.get(allocator_name)
                else:
                    self.logger.error(
                        "Allocator configuration not found in {}".format(allocator_cfg_file))
                    raise ResourceInitError(
                        "Allocator configuration not found in {}".format(allocator_cfg_file))
        except IOError as error:
            self.logger.error(error)
            raise ResourceInitError("Unable to read allocator config: {}".format(error))
        self.logger.debug("Read allocator configuration from {}: {}".format(allocator_cfg_file,
                                                                            allocator_config))
        return allocator_config



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'jusvat01'

from threading import Thread
import signal
import time


class Timeout(object):  # pylint: disable=too-few-public-methods
    """
    Timeout class using ALARM signal.
    """
    class Timeout(Exception):
        """
        Internal Timeout exception
        """
        pass

    def __init__(self, sec):
        self.sec = sec

    def __enter__(self):
        signal.signal(signal.SIGABRT, self.raise_timeout)
        signal.alarm(self.sec)

    def __exit__(self, *args):
        signal.alarm(0)    # disable alarm

    def raise_timeout(self, *args):  # pylint: disable=unused-argument,no-self-use
        """
        Raise a Timeout exception
        :param args: Not used
        :return: Nothing
        :raises: Timeout
        """
        raise Timeout.Timeout()


class Timer(Thread):
    """
    Timer class, simple timer that sleeps for timeout seconds.
    """
    def __init__(self):
        super(Timer, self).__init__()
        self.timeout = None

    def wait(self, timeout):
        """
        Starts the Timer for timeout seconds, then gives 5 second grace period to join the thread.
        :param timeout: Duration for timer.
        :return: Nothing
        """
        self.timeout = timeout
        self.start()
        self.join(timeout=5)

    def run(self):
        """
        Just sleep for self.timeout seconds.
        :return:
        """
        time.sleep(self.timeout)



# pylint: disable=too-many-instance-attributes

"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import platform
import signal
import subprocess
import time
import select
from threading import Thread, Lock
from collections import deque
from icetea_lib.TestStepError import TestStepError
from icetea_lib.tools.tools import strip_escape, is_pid_running, UNIXPLATFORM, IS_PYTHON3
import icetea_lib.LogManager as LogManager


class StreamDescriptor(object):  # pylint: disable=too-few-public-methods
    """
    StreamDescriptor class, container for stream components.
    """
    def __init__(self, stream, callback):
        self.stream = stream
        self.buf = ""
        self.read_queue = deque()  # pylint: disable=invalid-name
        self.has_error = False
        self.callback = callback


class NonBlockingStreamReader(object):
    """
    Implementation for a non-blocking stream reader.
    """
    _instance = None
    _streams = None
    _stream_mtx = None
    _rt = None
    _run_flag = False

    def __init__(self, stream, callback=None):
        # Global class variables
        if NonBlockingStreamReader._rt is None:
            NonBlockingStreamReader._streams = []
            NonBlockingStreamReader._stream_mtx = Lock()
            NonBlockingStreamReader._run_flag = True
            NonBlockingStreamReader._rt = Thread(target=NonBlockingStreamReader.run)
            NonBlockingStreamReader._rt.setDaemon(True)
            NonBlockingStreamReader._rt.start()
        # Instance variables
        self._descriptor = StreamDescriptor(stream, callback)

    def start(self):
        """
        Start the reader, acquires the global lock before appending the descriptor on the stream.
        Releases the lock afterwards.
        :return: Nothing
        """
        NonBlockingStreamReader._stream_mtx.acquire()
        NonBlockingStreamReader._streams.append(self._descriptor)
        NonBlockingStreamReader._stream_mtx.release()

    @staticmethod
    def _get_sd(file_descr):
        """
        Get streamdescriptor matching file_descr fileno.

        :param file_descr: file object
        :return: StreamDescriptor or None
        """
        for stream_descr in NonBlockingStreamReader._streams:
            if file_descr == stream_descr.stream.fileno():
                return stream_descr
        return None

    @staticmethod
    def _read_fd(file_descr):
        """
        Read incoming data from file handle.
        Then find the matching StreamDescriptor by file_descr value.

        :param file_descr: file object
        :return: Return number of bytes read
        """
        try:
            line = os.read(file_descr, 1024 * 1024)
        except OSError:
            stream_desc = NonBlockingStreamReader._get_sd(file_descr)
            if stream_desc is not None:
                stream_desc.has_error = True
                if stream_desc.callback is not None:
                    stream_desc.callback()
            return 0

        if line:
            stream_desc = NonBlockingStreamReader._get_sd(file_descr)
            if stream_desc is None:
                return 0 # Process closing

            if IS_PYTHON3:
                try:
                    # @TODO: further develop for not ascii/unicode binary content
                    line = line.decode("ascii")
                except UnicodeDecodeError:
                    line = repr(line)
            stream_desc.buf += line
            # Break lines
            split = stream_desc.buf.split(os.linesep)
            for line in split[:-1]:
                stream_desc.read_queue.appendleft(strip_escape(line.strip()))
                if stream_desc.callback is not None:
                    stream_desc.callback()
            # Store the remainded, its either '' if last char was '\n'
            # or remaining buffer before line end
            stream_desc.buf = split[-1]
            return len(line)
        return 0

    @staticmethod
    def _read_select_poll(poll):
        """
        Read PIPEs using select.poll() method
        Available on Linux and some Unixes
        """
        npipes = len(NonBlockingStreamReader._streams)
        for stream_descr in NonBlockingStreamReader._streams:
            if not stream_descr.has_error:
                poll.register(stream_descr.stream,
                              select.POLLIN | select.POLLERR | select.POLLHUP | select.POLLNVAL)

        while NonBlockingStreamReader._run_flag:
            for (file_descr, event) in poll.poll(500):
                if event == select.POLLIN:
                    NonBlockingStreamReader._read_fd(file_descr)
                else:
                    # Because event != select.POLLIN, the pipe is closed
                    # but we still want to read all bytes
                    while NonBlockingStreamReader._read_fd(file_descr) != 0:
                        pass
                    # Dut died, signal the processing thread so it notices that no lines coming in
                    stream_descr = NonBlockingStreamReader._get_sd(file_descr)
                    if stream_descr is None:
                        return # PIPE closed but DUT already disappeared
                    stream_descr.has_error = True
                    if stream_descr.callback is not None:
                        stream_descr.callback()
                        return # Force poll object to reregister only alive descriptors

            # Check if new pipes added, don't need mutext just for reading the size
            # If we will not get it right now, we will at next time
            if npipes != len(NonBlockingStreamReader._streams):
                return

    @staticmethod
    def _read_select_kqueue(k_queue):
        """
        Read PIPES using BSD Kqueue
        """
        npipes = len(NonBlockingStreamReader._streams)
        # Create list of kevent objects
        # pylint: disable=no-member
        kevents = [select.kevent(s.stream.fileno(),
                                 filter=select.KQ_FILTER_READ,
                                 flags=select.KQ_EV_ADD | select.KQ_EV_ENABLE)
                   for s in NonBlockingStreamReader._streams]
        while NonBlockingStreamReader._run_flag:
            events = k_queue.control(kevents, npipes, 0.5)  # Wake up twice in second
            for event in events:
                if event.filter == select.KQ_FILTER_READ:  # pylint: disable=no-member
                    NonBlockingStreamReader._read_fd(event.ident)
            # Check if new pipes added.
            if npipes != len(NonBlockingStreamReader._streams):
                return

    @staticmethod
    def run():
        """
        Run loop
        """
        while NonBlockingStreamReader._run_flag:
            # Wait for streams to appear
            if not NonBlockingStreamReader._streams:
                time.sleep(0.2)
                continue
            # Try to get correct select/poll method for this OS
            # Try if select.poll() is supported (Linux/UNIX)
            try:
                poll = select.poll()
            except AttributeError:
                pass
            else:
                NonBlockingStreamReader._read_select_poll(poll)
                del poll
                continue
            # Try is select.kqueue is supported (BSD/OS X)
            try:
                k_queue = select.kqueue()  # pylint: disable=no-member
            except AttributeError:
                pass
            else:
                NonBlockingStreamReader._read_select_kqueue(k_queue)
                k_queue.close()
                continue
            # Not workable polling method found
            raise RuntimeError('This OS is not supporting select.poll() or select.kqueue()')

    def stop(self):
        """
        Stop the reader
        """
        # print('stopping NonBlockingStreamReader..')
        # print('acquire..')
        NonBlockingStreamReader._stream_mtx.acquire()
        # print('acquire..ok')
        NonBlockingStreamReader._streams.remove(self._descriptor)
        if not NonBlockingStreamReader._streams:
            NonBlockingStreamReader._run_flag = False
        # print('release..')
        NonBlockingStreamReader._stream_mtx.release()
        # print('release..ok')
        if NonBlockingStreamReader._run_flag is False:
            # print('join..')
            NonBlockingStreamReader._rt.join()
            # print('join..ok')
            del NonBlockingStreamReader._rt
            NonBlockingStreamReader._rt = None
            # print('stopping NonBlockingStreamReader..ok')

    def has_error(self):
        """
        :return: Boolean, True if _descriptor.has_error is True. False otherwise
        """
        return self._descriptor.has_error

    def readline(self):
        """
        Readline implementation.

        :return: popped line from descriptor queue. None if nothing found
        :raises: RuntimeError if errors happened while reading PIPE
        """
        try:
            return self._descriptor.read_queue.pop()
        except IndexError:
            # No lines in queue
            if self.has_error():
                raise RuntimeError("Errors reading PIPE")
        return None


class GenericProcess(object):
    """
    Generic process implementation for use with Dut.
    """
    # Contstruct GenericProcess instance
    def __init__(self, name, cmd=None, path=None, logger=None):
        self.name = name
        self.proc = None
        self.logger = logger
        self.cmd = None
        self.cmd_arr = None
        self.path = None
        self.gdb = False
        self.gdbs = False
        self.vgdb = False
        self.gdbs_port = None
        self.nobuf = False
        self.valgrind = None
        self.valgrind_xml = None
        self.valgrind_console = None
        self.valgrind_track_origins = None
        self.valgrind_extra_params = None
        self.__print_io = True
        self.__valgrind_log_basename = None
        self.read_thread = None
        self.__ignore_return_code = False
        self.default_retcode = 0

        if not self.logger:
            self.logger = LogManager.get_bench_logger(name, 'GP', False)
        self.cmd = cmd
        self.path = path

    def enable_io_prints(self):
        """
        Enable IO prints
        """
        self.__print_io = True

    def disable_io_prints(self):
        """
        Disable IO prints
        """
        self.__print_io = False

    @property
    def ignore_return_code(self):
        """
        Return value of __ignoreReturnCode
        """
        return self.__ignore_return_code

    @ignore_return_code.setter
    def ignore_return_code(self, value):
        """
        Set __ignoreReturnCode
        """
        self.__ignore_return_code = value

    # use gdb for process
    def use_gdb(self, gdb=True):
        """
        Set gdb use for process.

        :param gdb: Boolean, defaults to True.
        """
        self.gdb = gdb

    def use_gdbs(self, gdbs=True, port=2345):
        """
        Set gdbs use for process.

        :param gdbs: Boolean, default is True
        :param port: Port number for gdbserver
        """
        self.gdbs = gdbs
        self.gdbs_port = port

    # use vgdb for process
    def use_vgdb(self, vgdb=True):
        """
        Set vgdb for process.

        :param vgdb: Boolean, defaults to True
        """
        self.vgdb = vgdb

    def no_std_buf(self, nobuf=True):
        """
        Set buffering of stdio.

        :param nobuf: Defaults to True (no buffering)
        """
        self.nobuf = nobuf

    # pylint: disable=too-many-arguments
    def use_valgrind(self, tool, xml, console, track_origins, valgrind_extra_params):
        """
        Use Valgrind.

        :param tool: Tool name, must be memcheck, callgrind or massif
        :param xml: Boolean output xml
        :param console: Dump output to console, Boolean
        :param track_origins: Boolean, set --track-origins=yes
        :param valgrind_extra_params:  Extra parameters
        :return: Nothing
        :raises: AttributeError if invalid tool set.
        """
        self.valgrind = tool
        self.valgrind_xml = xml
        self.valgrind_console = console
        self.valgrind_track_origins = track_origins
        self.valgrind_extra_params = valgrind_extra_params
        if not tool in ['memcheck', 'callgrind', 'massif']:
            raise AttributeError("Invalid valgrind tool: %s" % tool)

    def __get_valgrind_params(self):
        """
        Get Valgrind command as list.

        :return: list
        """
        valgrind = []
        if self.valgrind:
            valgrind.extend(['valgrind'])
            if self.valgrind == 'memcheck':
                valgrind.extend(['--tool=memcheck', '--leak-check=full'])
                if self.valgrind_track_origins:
                    valgrind.extend(['--track-origins=yes'])
                if self.valgrind_console:
                    # just dump the default output, which is text dumped to console
                    valgrind.extend([])
                elif self.valgrind_xml:
                    valgrind.extend([
                        '--xml=yes',
                        '--xml-file=' + LogManager.get_testcase_logfilename(
                            self.name + '_valgrind_mem.xml', prepend_tc_name=True)
                    ])
                else:
                    valgrind.extend([
                        '--log-file=' + LogManager.get_testcase_logfilename(
                            self.name + '_valgrind_mem.txt')
                    ])

            elif self.valgrind == 'callgrind':
                valgrind.extend([
                    '--tool=callgrind',
                    '--dump-instr=yes',
                    '--simulate-cache=yes',
                    '--collect-jumps=yes'])
                if self.valgrind_console:
                    # just dump the default output, which is text dumped to console
                    valgrind.extend([])
                elif self.valgrind_xml:
                    valgrind.extend([
                        '--xml=yes',
                        '--xml-file=' + LogManager.get_testcase_logfilename(
                            self.name + '_valgrind_calls.xml', prepend_tc_name=True)
                    ])
                else:
                    valgrind.extend([
                        '--callgrind-out-file=' + LogManager.get_testcase_logfilename(
                            self.name + '_valgrind_calls.data')
                    ])
            elif self.valgrind == 'massif':
                valgrind.extend(['--tool=massif'])
                valgrind.extend([
                    '--massif-out-file=' + LogManager.get_testcase_logfilename(
                        self.name + '_valgrind_massif.data')
                    ])
            # this allows one to specify misc params to valgrind,
            # eg. "--threshold=0.4" to get some more data from massif
            if self.valgrind_extra_params != '':
                valgrind.extend(self.valgrind_extra_params.split())

        return valgrind

    def start_process(self, cmd=None, path="", processing_callback=None):
        """
        Start the process.

        :param cmd: Command to run
        :param path: cwd
        :param processing_callback: Callback for processing lines
        :return: Nothing
        :raises: NameError if Connection fails
        """
        self.cmd = self.cmd if not cmd else cmd
        self.path = self.path if not path else path
        if self.path:
            self.path = os.path.abspath(self.path)
        self.cmd_arr = []

        # set stdbuf in/out/err to zero size = no buffers in use
        if self.nobuf:
            self.cmd_arr.extend(['stdbuf', '-i0', '-o0', '-e0'])

        # check if user want to debug this process
        if self.gdb:
            # add gdb parameters, run program immediately
            self.cmd_arr.extend(['gdb', '-ex=run', '--args'])
        elif self.gdbs:
            # add gdbserver parameters, run program immediately
            self.cmd_arr.extend(['gdbserver', 'localhost:' + str(self.gdbs_port)])
        elif self.vgdb:
            # add valgrind vgdb parameters, run program but wait for remote gdb connection
            self.cmd_arr.extend(['valgrind', '--vgdb=yes', '--vgdb-error=0'])

        if self.valgrind:
            self.cmd_arr.extend(self.__get_valgrind_params())

        self.cmd_arr.extend(self.cmd)
        prefn = None
        if not platform.system() == "Windows":
            prefn = os.setsid

        self.logger.debug("Instantiating process "
                          "%s at %s with command %s"
                          % (self.name, self.path, " ".join(self.cmd_arr)),
                          extra={"type": "   "})
        self.proc = subprocess.Popen(self.cmd_arr, cwd=self.path, stdout=subprocess.PIPE,
                                     stdin=subprocess.PIPE, preexec_fn=prefn)

        if UNIXPLATFORM:
            import fcntl
            file_descr = self.proc.stdout.fileno()
            fcntl_var = fcntl.fcntl(file_descr, fcntl.F_GETFL)
            fcntl.fcntl(file_descr, fcntl.F_SETFL, fcntl_var | os.O_NONBLOCK)

        if self.proc.pid:
            # Start stream reader thread
            self.read_thread = NonBlockingStreamReader(self.proc.stdout, processing_callback)
            self.read_thread.start()
            self.logger.info("Process '%s' running with pid: %i" % (' '.join(self.cmd_arr),
                                                                    self.proc.pid),
                             extra={'type': '<->'})
        else:
            self.logger.warning("Process start fails", extra={'type': '<->'})
            raise NameError('Connection Fails')

    def stop_process(self):
        """
        Stop the process.

        :raises: EnvironmentError if stopping fails due to unknown environment
        TestStepError if process stops with non-default returncode and return code is not ignored.
        """
        if self.read_thread is not None:
            self.logger.info("stop_process::readThread.stop()-in")
            self.read_thread.stop()
            self.logger.info("stop_process::readThread.stop()-out")
        returncode = None
        if self.proc:
            self.logger.info("os.killpg(%d)", self.proc.pid)
            try:
                try:
                    os.killpg(self.proc.pid, signal.SIGTERM)
                except AttributeError:
                    self.logger.info("os.killpg::AttributeError")
                    # Failed most likely because in windows,
                    # so use taskkill to kill whole process tree of proc
                    if platform.system() == "Windows":
                        subprocess.call(['taskkill', '/F', '/T', '/PID', str(self.proc.pid)])
                    else:
                        self.logger.info("os.killpg::unknown env")
                        raise EnvironmentError("Unknown platform, "
                                               "don't know how to terminate process")
                self.proc.communicate()  # Wait for pipes to clear and process to stop.
                returncode = self.proc.wait()
            except OSError:
                self.logger.info("os.killpg::OSError")
            del self.proc

        if returncode is not None:
            self.logger.info("stop_process::returncode is not None:")
            self.logger.info("Process stopped with returncode %s" % returncode)
            if returncode != self.default_retcode and not self.__ignore_return_code:
                raise TestStepError("Process stopped with returncode %d" % returncode)
        self.logger.info("stop_process-out")

    def stop(self):
        """
        Stop the process
        See stop_process for more information
        """
        self.stop_process()

    def readline(self, timeout=1):  # pylint: disable=unused-argument
        """
        Readline implementation.

        :param timeout: Timeout, not used
        :return: Line read or None
        """
        data = None
        if self.read_thread:
            # Ignore the timeout value, return imediately if no lines in queue
            data = self.read_thread.readline()
            if data and self.__print_io:
                self.logger.info(data, extra={'type': '<--'})
        return data

    def writeline(self, data, crlf="\r\n"):
        """
        Writeline implementation.

        :param data: Data to write
        :param crlf: Line end characters, defailt is \r\n
        :return: Nothing
        :raises: RuntimeError if errors happen while writing to PIPE or process stops.
        """
        if self.read_thread:
            if self.read_thread.has_error():
                raise RuntimeError("Error writing PIPE")
        # Check if process still alive
        if self.proc.poll() is not None:
            raise RuntimeError("Process stopped")
        if self.__print_io:
            self.logger.info(data, extra={'type': '-->'})
        self.proc.stdin.write(bytearray(data + crlf, 'ascii'))
        self.proc.stdin.flush()

    def is_alive(self):
        """
        Is process alive.

        :return: Boolean, True is process is still running.
        """
        return is_pid_running(self.proc.pid) if self.proc else False



# pylint: disable=missing-docstring,expression-not-assigned

"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import unittest

from icetea_lib.LogManager import ContextFilter, traverse_json_obj
from icetea_lib.tools.tools import IS_PYTHON3


class ContextFilterTest(unittest.TestCase):

    # Helper function
    @staticmethod
    def create_log_record(msg):
        return logging.LogRecord(name="", level=logging.ERROR, pathname="",
                                 lineno=0, msg=msg, args=None, exc_info=None,
                                 func=None)

    def setUp(self):
        self.contextfilter = ContextFilter()

    def test_filter_base64(self):
        msg = "aaa="
        record = self.create_log_record(msg)
        self.contextfilter.filter(record)
        self.assertEqual(msg, record.msg)

        msg = []
        [msg.append("a") for _ in range(ContextFilter.MAXIMUM_LENGTH)]
        msg = "".join(msg)
        record = self.create_log_record(msg)
        self.contextfilter.filter(record)
        expected = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...[9950 more bytes]"
        self.assertEqual(expected, record.msg)

        msg = []
        [msg.append("a") for _ in range(ContextFilter.MAXIMUM_LENGTH * 2)]
        msg = "".join(msg)
        record = self.create_log_record(msg)
        self.contextfilter.filter(record)
        expected = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...[19950 more bytes]"
        self.assertEqual(expected, record.msg)

        msg = []
        [msg.append("a") for _ in range(ContextFilter.MAXIMUM_LENGTH - 1)]
        msg = "".join(msg)
        record = self.create_log_record(msg)
        self.contextfilter.filter(record)
        self.assertEqual(msg, record.msg)

        msg = []
        [msg.append("a") for _ in range(ContextFilter.MAXIMUM_LENGTH + 1)]
        msg = "".join(msg)
        record = self.create_log_record(msg)
        self.contextfilter.filter(record)
        expected = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...[9951 more bytes]"
        self.assertEqual(expected, record.msg)

    def test_filter_human_readable(self):
        msg = [" "]
        [msg.append("a") for _ in range(ContextFilter.MAXIMUM_LENGTH)]
        msg = "".join(msg)
        record = self.create_log_record(msg)
        expected = " aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa...[9951 more bytes]"
        self.contextfilter.filter(record)
        self.assertEqual(expected, record.msg)

    def test_filter_binary_data(self):
        msg = []
        [msg.append(os.urandom(1024)) for _ in range(ContextFilter.MAXIMUM_LENGTH +1)]
        msg = b"".join(msg)
        expected = "{}...[10240974 more bytes]".format(msg[:50])
        record = self.create_log_record(msg)
        self.contextfilter.filter(record)

        self.assertEqual(expected, record.msg)

        msg = []
        [msg.append(u'\ufffd') for _ in range(ContextFilter.MAXIMUM_LENGTH + 1)]
        msg = "".join(msg)
        if not IS_PYTHON3:
            expected = u"{}...[9951 more bytes]".format(repr(msg[:50]))
        else:
            expected = u"{}...[9951 more bytes]".format(msg[:50])

        record = self.create_log_record(msg)
        self.contextfilter.filter(record)
        self.assertEqual(expected, record.msg)


class TraverseJsonObjTest(unittest.TestCase):
    def test_stays_untouched(self):
        test_dict = [{"a": "aa", "b": ["c", "d", {"e": "aa", "f": "aa"}]}]

        self.assertEqual(test_dict, traverse_json_obj(test_dict))

    def test_all_modified(self):
        test_dict = [{"a": "aa", "b": ["c", "d", {"e": "aa", "f": "aa"}]}]

        def modify(value):
            return "bb" if value == "aa" else value

        expected = [{"a": "bb", "b": ["c", "d", {"e": "bb", "f": "bb"}]}]
        self.assertEqual(expected, traverse_json_obj(test_dict, callback=modify))



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Dut module. Dut is the base class for all different Dut types and many of the lower level
functionalities of duts are implemented here, including the read thread run loop and signaling
mechanisms.
"""


import re
import time
import types
from collections import deque
from threading import Thread, Event, Semaphore
from six import string_types

import icetea_lib.LogManager as LogManager
from icetea_lib.CliAsyncResponse import CliAsyncResponse
from icetea_lib.CliRequest import CliRequest
from icetea_lib.CliResponse import CliResponse
from icetea_lib.Events.EventMatcher import EventMatcher
from icetea_lib.Events.Generics import EventTypes
from icetea_lib.Events.Generics import Event as EventObject
from icetea_lib.TestStepError import TestStepError, TestStepFail, TestStepTimeout
from icetea_lib.tools.tools import num


class DutConnectionError(Exception):
    """
    Exception for errors in connecting to dut.
    """
    pass


class Location(object):  # pylint: disable=too-few-public-methods
    """
    Location object for storing x and y coordinates of a dut in a network topology.
    """
    def __init__(self, x_coord=0, y_coord=0):
        self.x_coord = x_coord
        self.y_coord = y_coord


class Dut(object):
    """
    The Dut object is the base behind all our Dut types. It implements the read thread run loop
    and Dut signaling mechanisms and contains many of the shared attributes, methods and data of
    it's subclasses.
    """

    _dutlist = None
    _run = False
    _th = None
    _logger = None
    _sem = None
    _signalled_duts = None

    def __init__(self, name, params=None):
        if Dut._dutlist is None:
            Dut._dutlist = []
        self.testcase = ''
        self.version = {}
        self.dutinformation = None
        self.name = name
        self.dut_name = name
        self.stopped = False
        self.comport = False  # TODO: Move to DutProcess?
        # TODO: Move to MeshCommands
        self.MAC = None  # pylint: disable=invalid-name
        self.location = Location(0, 0)
        self.logger = LogManager.get_bench_logger('Dut.%s' % name, name)
        self.get_time = time.time
        self.query = None # Query to node
        self.query_timeout = 0
        self.query_async_expected = None  # Expected retcode for async cmd
        self.query_async_response = None  # Async response to fullfill when response is available
        self.waiting_for_response = None
        self.response_coming_in = None  # Response coming in
        self.prev = None  # Previous command, stored for logging purposes
        self._store_traces = True
        self.traces = []  # All traces
        self.response_traces = []  # Incoming response lines
        self.response_received = Event()
        self.response_received.set()
        self.config = {}
        self.init_cli_cmds = None
        self.post_cli_cmds = None
        self.params = params
        self.index = None
        self.init_done = Event()
        self.init_event_matcher = None
        self.init_wait_timeout = None
        Dut._dutlist.append(self)

    @property
    def platform(self):
        """
        Getter for dut platform.

        :return: Platform as string or None
        """
        if self.dutinformation:
            return self.dutinformation.platform
        return None

    @platform.setter
    def platform(self, value):
        if self.dutinformation:
            self.dutinformation.platform = value

    @property
    def index(self):
        """
        Getter for dut index.

        :return: Index as integer or None
        """
        if self.dutinformation:
            return self.dutinformation.index
        return None

    @index.setter
    def index(self, value):
        if self.dutinformation:
            self.dutinformation.index = value

    @property
    def resource_id(self):
        """
        Getter for dut resource id.

        :return: Resource id as string or None
        """
        if self.dutinformation:
            return self.dutinformation.resource_id
        return None

    @resource_id.setter
    def resource_id(self, value):
        if self.dutinformation:
            self.dutinformation.resource_id = value

    @property
    def vendor(self):
        """
        Getter for dut vendor.

        :return: Vendor as string. If not found, will be empty string.
        """
        if self.dutinformation:
            return self.dutinformation.vendor

    @vendor.setter
    def vendor(self, value):
        if self.dutinformation:
            self.dutinformation.vendor = value

    @property
    def build(self):
        """
        Getter for build information.

        :return: Build object or None
        """
        if self.dutinformation:
            return self.dutinformation.build

    @build.setter
    def build(self, value):
        if self.dutinformation:
            self.dutinformation.build = value

    @property
    def store_traces(self):
        """
        Getter for _store_traces. _store_traces controls in memory storing of received lines.

        :return: Boolean
        """
        return self._store_traces

    @store_traces.setter
    def store_traces(self, value):
        """
        Setter for _store_traces. _store_traces controls in memory storing of received lines.
        Also logs the change for the user.

        :param value: Boolean
        :return: Nothing
        """
        if not value:
            self.logger.debug("Stopping storing received lines for dut %d", self.index)
            self._store_traces = False
        else:
            self.logger.debug("Resuming storing received lines for dut %d", self.index)
            self._store_traces = True

    # Minimum requirements from Dut Implementation
    def open_connection(self):
        """
        Open connection to DUT.
        Connection information is stored in variable self.comport
        """
        raise NotImplementedError("open_connection is not implemented")

    def close_connection(self):
        """
        Close connection to DUT
        """
        raise NotImplementedError("close_connection is not implemented")

    def writeline(self, data):
        """
        write single string line to DUT

        :param data: string to be write to DUT, not contains line feeds,
        instead writeline implementation should add line feeds when needed.
        """
        raise NotImplementedError("writeline is not implemented")

    def readline(self, timeout=1):
        """read single string line from DUT.

        :param timeout: readline timeout in seconds.
        :return string from DUT.
        """
        raise NotImplementedError("readline is not implemented")

    def print_info(self):
        '''
        Log information relevant to this DUT.

        :return: Nothing
        '''
        raise NotImplementedError("printInfo has not been implemented")

    def get_info(self):
        """
        Creates a DutInformation object that contains information on this DUT.

        :return: DutInformation object
        """
        raise NotImplementedError("getInfo has not been implemented")

    # Minimum requirements from Dut Implementation - END

    def set_log_level(self, level):
        """
        Set the level of logging for this dut. NOTE: AFFECTS FILES AS WELL

        :param level: Level to set (integer or constant from logging)
        :return: Nothing
        """
        self.logger.setLevel(level)

    def set_dut_name(self, name):
        """
        Set dut name

        :param name: name to set
        :return: Nothing
        """
        self.name = name
        self.dut_name = name

    def _flash_needed(self, **kwargs):
        """
        Check if flashing of dut is required.

        :return: Boolean. True if flashing is needed, else False
        """
        raise NotImplementedError("flashing needed check not implemented!")

    def get_dut_name(self):
        """
        Get dut dname

        :return: dut name
        """
        return self.dut_name

    def set_init_cli_cmds(self, cmds):
        """
        set cli initialization commands

        :param cmds: list of commands as either strings or lists of strings
        :return: Nothing
        """
        self.init_cli_cmds = cmds

    def set_post_cli_cmds(self, cmds):
        """
        set command to restore cli status to state before init_cli_cmds
        :param cmds: list of commands as either strings or lists of strings
        :return: Nothing
        """
        self.post_cli_cmds = cmds

    def set_default_init_cli_cmds(self):
        """
        Default init commands are set --retcode true, echo off, set --vt100 off, set dut <dut name>
        and set testcase <tc name>

        :return: List of default cli initialization commands.
        """
        init_cli_cmds = []
        init_cli_cmds.append("set --retcode true")
        init_cli_cmds.append("echo off")
        init_cli_cmds.append("set --vt100 off")

        #set dut name as variable
        init_cli_cmds.append('set dut "'+self.name+'"')
        init_cli_cmds.append(['set testcase "' + self.testcase + '"', True])
        return init_cli_cmds

    def set_default_init_cli_human_cmds(self):  # pylint: disable=no-self-use
        """
        Default commands to restore cli to human readable state are echo on, set --vt100 on,
        set --retcode false.

        :return: List of default commands to restore cli to human readable format
        """
        post_cli_cmds = []
        post_cli_cmds.append("echo on")
        post_cli_cmds.append("set --vt100 on")
        post_cli_cmds.append(["set --retcode false", False, False])  # last True is wait=<Boolean>
        return post_cli_cmds

    def init_cli(self):
        """
        Init cli for script.
        '
        :return: Nothing
        """
        if self.init_cli_cmds is None:
            self.init_cli_cmds = self.set_default_init_cli_cmds()

        for cli_cmd in self.init_cli_cmds:
            if isinstance(cli_cmd, list) and len(cli_cmd) >= 2:
                asynchronous = cli_cmd[1]
                if len(cli_cmd) > 2:
                    wait = cli_cmd[2]
                else:
                    wait = True
                self.execute_command(cli_cmd[0], wait=wait, asynchronous=asynchronous)
            else:
                self.execute_command(cli_cmd)

    def init_wait_register(self):
        """
        Initialize EventMatcher to wait for certain cli_ready_trigger to arrive from this Dut.

        :return: None
        """
        app = self.config.get("application")

        if app:
            bef_init_cmds = app.get("cli_ready_trigger")
            if bef_init_cmds:
                self.init_done.clear()
                self.init_event_matcher = EventMatcher(EventTypes.DUT_LINE_RECEIVED,
                                                       bef_init_cmds, self, self.init_done)
                self.init_wait_timeout = app.get("cli_ready_trigger_timeout", 30)
                return
        self.init_done.set()
        return

    def wait_init(self):
        """
        Block until init_done flag is set or until init_wait_timeout happens.

        :return: value of init_done
        """
        init_done = self.init_done.wait(timeout=self.init_wait_timeout)
        if not init_done:
            if hasattr(self, "peek"):
                app = self.config.get("application")
                if app:
                    bef_init_cmds = app.get("cli_ready_trigger")
                    if bef_init_cmds in self.peek():  # pylint: disable=no-member
                        init_done = True
        return init_done

    def init_cli_human(self):
        """
        Send post_cli_cmds to dut

        :return: Nothing
        """
        if self.post_cli_cmds is None:
            self.post_cli_cmds = self.set_default_init_cli_human_cmds()
        for cli_cmd in self.post_cli_cmds:
            try:
                if isinstance(cli_cmd, list) and len(cli_cmd) >= 2:
                    asynchronous = cli_cmd[1]
                    if len(cli_cmd) > 2:
                        wait = cli_cmd[2]
                    else:
                        wait = True
                    self.execute_command(cli_cmd[0], wait=wait, asynchronous=asynchronous)
                else:
                    self.execute_command(cli_cmd)
            except (TestStepFail, TestStepError, TestStepTimeout):
                continue

    def set_time_function(self, function):
        """
        Set time function to be used.

        :param function: callable function
        :return: Nothing
        :raises: ValueError if function is not types.FunctionType.
        """
        if isinstance(function, types.FunctionType):
            self.get_time = function
        else:
            raise ValueError("Invalid value for DUT time function")

    def open_dut(self, port=None):
        """
        Open connection to dut.

        :param port: com port to use.
        :return:
        """
        if port is not None:
            self.comport = port

        try:
            self.open_connection()
        except (DutConnectionError, ValueError) as err:
            self.close_dut(use_prepare=False)
            raise DutConnectionError(str(err))
        except KeyboardInterrupt:
            self.close_dut(use_prepare=False)
            self.close_connection()
            raise

    def _wait_for_exec_ready(self):
        """
        Wait for response.

        :return: CliResponse object coming in
        :raises: TestStepTimeout, TestStepError
        """
        #wait for response
        while not self.response_received.wait(1) and self.query_timeout != 0:
            if self.query_timeout != 0 and self.query_timeout < self.get_time():
                if self.prev:
                    cmd = self.prev.cmd
                else:
                    cmd = "???"
                self.logger.error("CMD timeout: "+ cmd)
                self.query_timeout = 0
                raise TestStepTimeout(self.name + " CMD timeout: " + cmd)
            self.logger.debug("Waiting for response... "
                              "timeout=%d", self.query_timeout - self.get_time())

        if self.response_coming_in == -1:
            if self.query_async_response is not None:
                # fullfill the async response with a dummy response and clean the state
                self.query_async_response.set_response(CliResponse())
                self.query_async_response = None
            # raise and log the error
            self.logger.error("No response received, DUT died")
            raise TestStepError("No response received, DUT "+self.name+" died")

        # if an async response is pending, fullfill it with the result
        if self.query_async_response is not None:
            self.query_async_response.set_response(self.response_coming_in)
            self.query_async_response = None

        self.query_timeout = 0
        return self.response_coming_in

    def execute_command(self, req, **kwargs):
        """
        Execute command and return CliResponse

        :param req: String, command to be executed in DUT, or CliRequest, command class which
        contains all configurations like timeout.
        :param kwargs: Configurations (wait, timeout) which will be used when string mode is in use.
        :return: CliResponse, which contains all received data from Dut and parsed retcode.
        """
        if isinstance(req, string_types):
            # backward compatible
            timeout = 50 # Use same default timeout as bench.py
            wait = True
            asynchronous = False
            for key in kwargs:
                if key == 'wait':
                    wait = kwargs[key]
                elif key == 'timeout':
                    timeout = kwargs[key]  # [ms]
                elif key == 'asynchronous':
                    asynchronous = kwargs[key]
            req = CliRequest(req,
                             timestamp=self.get_time(),
                             wait=wait,
                             timeout=timeout,
                             asynchronous=asynchronous)

        # wait for previous command ready
        if req.wait:
            response = self._wait_for_exec_ready()
            if response is not None and self.query_async_expected is not None:
                if response.retcode != self.query_async_expected:
                    self.logger.error("Asynch call returned unexpected result, "
                                      "expected %d was %d", self.query_async_expected,
                                      response.retcode)
                    raise TestStepFail("Asynch call returned unexpected result")
                self.query_async_expected = None

        # Tell Query to worker thread
        self.response_received.clear()
        self.query_timeout = self.get_time() + req.timeout if req.wait else 0
        self.query = req

        msg = "Async CMD {}, timeout={}, time={}" if req.asynchronous else "CMD {}, timeout={}, time={}"
        msg = msg.format(req.cmd, int(self.query_timeout), int(self.get_time()))
        self.logger.debug(msg, extra={'type': '<->'})

        Dut.process_dut(self)

        if req.asynchronous is True:
            self.query_async_expected = req.expected_retcode
            async_response = CliAsyncResponse(self)
            self.query_async_response = async_response
            return async_response

        if req.wait is False:
            self.query_async_expected = req.expected_retcode
            # if an async response was waiting, just discard the result
            # since the new command has already been sent...
            # This is not ideal but when a command has its flags "Wait == False"
            # the result of the previous command is already discarded in previous
            # stages
            if self.query_async_response is not None:
                self.query_async_response.set_response(CliResponse())
            self.query_async_response = None
            return CliResponse()

        return self._wait_for_exec_ready()

    def close_dut(self, use_prepare=True):
        """
        Close connection to dut.

        :param use_prepare: Boolean, default is True. Call prepare_connection_close before
        closing connection.
        :return: Nothing
        """
        if not self.stopped:
            self.logger.debug("Close '%s' connection" % self.dut_name, extra={'type': '<->'})
            if use_prepare:
                try:
                    self.prepare_connection_close()
                except TestStepFail:
                    # We can ignore this for dead Duts, just continue with cleanup
                    pass
            self.stopped = True
            Dut._dutlist.remove(self)
            # Remove myself from signalled dut list, if I'm still there
            if Dut._signalled_duts and Dut._signalled_duts.count(self):
                try:
                    Dut._signalled_duts.remove(self)
                except ValueError:
                    pass

            try:
                if not Dut._dutlist:
                    Dut._run = False
                    Dut._sem.release()
                    Dut._th.join()
                    del Dut._th
                    Dut._th = None
            except AttributeError:
                pass

    def reset(self, method=None):  # pylint: disable=unused-argument
        """
        Reset the dut. Some duts implement different reset methods (hard/soft)

        :param method: String (hard or soft) or None
        :return: Nothing
        :raises: DutConnectionError if reset fails or connection is not restored.
        """
        raise NotImplementedError("reset function not implemented")

    def prepare_connection_close(self):
        """
        Run commands or other required things to prepare to close dut connection.

        :return: Nothing
        """
        pass

    @staticmethod
    def process_dut(dut):
        """
        Signal worker thread that specified Dut needs processing
        """
        if dut.finished():
            return
        Dut._signalled_duts.appendleft(dut)
        Dut._sem.release()

    # Thread runner
    @staticmethod
    def run():
        """
        Main thread runner for all Duts.

        :return: Nothing
        """
        Dut._logger.debug("Start DUT communication", extra={'type': '<->'})
        while Dut._run:
            Dut._sem.acquire()
            try:
                dut = Dut._signalled_duts.pop()
                # Check for pending requests
                if dut.waiting_for_response is not None:
                    item = dut.waiting_for_response
                    # pylint: disable=protected-access
                    dut.response_coming_in = dut._read_response()
                    if dut.response_coming_in is None:
                        # Continue to next node
                        continue
                    if isinstance(dut.response_coming_in, CliResponse):
                        dut.response_coming_in.set_response_time(item.get_timedelta(dut.get_time()))
                    dut.waiting_for_response = None
                    dut.logger.debug("Got response", extra={'type': '<->'})
                    dut.response_received.set()
                    continue

                # Check for new Request
                if dut.query is not None:
                    item = dut.query
                    dut.query = None
                    dut.logger.info(item.cmd, extra={'type': '-->'})
                    try:
                        dut.writeline(item.cmd)
                    except RuntimeError:
                        dut.response_coming_in = -1
                        dut.response_received.set()
                        continue
                    dut.prev = item # Save previous command for logging purposes
                    if item.wait:
                        # Only caller will care if this was asynchronous.
                        dut.waiting_for_response = item
                    else:
                        dut.query_timeout = 0
                        dut.response_received.set()
                    continue

                try:
                    line = dut.readline()
                except RuntimeError:
                    dut.response_coming_in = -1
                    dut.response_received.set()
                    continue
                if line:
                    if dut.store_traces:
                        dut.traces.append(line)
                    EventObject(EventTypes.DUT_LINE_RECEIVED, dut, line)
                    retcode = dut.check_retcode(line)
                    if retcode is not None:
                        dut.logger.warning("unrequested retcode", extra={'type': '!<-'})
                    dut.logger.debug(line, extra={'type': '<<<'})

            except IndexError:
                pass
        Dut._logger.debug("End DUT communication", extra={'type': '<->'})

    def _read_response(self):
        """
        Internal response reader.

        :return: CliResponse or None
        """
        try:
            line = self.readline()
        except RuntimeError:
            Dut._logger.warning("Failed to read PIPE", extra={'type': '!<-'})
            return -1
        if line:
            if self.store_traces:
                self.traces.append(line)
                self.response_traces.append(line)
            EventObject(EventTypes.DUT_LINE_RECEIVED, self, line)
            match = re.search(r"^\[([\w\W]{4})\]\[([\W\w]{4,}?)\]\: (.*)", line)
            if match:
                self.logger.debug(line, extra={'type': '<<<'})
            else:
                self.logger.info(line, extra={'type': '<--'})

            retcode = self.check_retcode(line)
            if retcode is not None:
                resp = CliResponse()
                resp.retcode = retcode
                resp.traces = self.response_traces
                resp.lines = self.response_traces
                self.response_traces = []
                return resp
        return None

    # check retcode
    def check_retcode(self, line):
        """
        Look for retcode on line line and return return code if found.

        :param line: Line to search from
        :return: integer return code or -1 if "cmd tasklet init" is found. None if retcode or cmd
        tasklet init not found.
        """
        retcode = None
        match = re.search(r"retcode\: ([-\d]{1,})", line)
        if match:
            retcode = num(str(match.group(1)))

        match = re.search("cmd tasklet init", line)
        if match:
            self.logger.debug("Device Boot up", extra={'type': '   '})
            return -1
        return retcode

    def finished(self):
        """
        Getter for stopped

        :return: Boolean
        """
        return self.stopped

    def start_dut_thread(self):  # pylint: disable=no-self-use
        """
        Start Dut thread.

        :return: Nothing
        """
        if Dut._th is None:
            Dut._run = True
            Dut._sem = Semaphore(0)
            Dut._signalled_duts = deque()
            Dut._logger = LogManager.get_bench_logger('Dut')
            Dut._th = Thread(target=Dut.run, name='DutThread')

            Dut._th.daemon = True
            Dut._th.start()



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
import mock
import threading
import logging


from icetea_lib.bench import Bench, ReturnCodes
from icetea_lib.TestStepError import TestStepFail, TestStepError, InconclusiveError, TestStepTimeout
from test.tests.test_tcTearDown import Testcase as TearDownTest


class MockDut:
    def __init__(self):
        self.traces = ["this is test line 1", "this is test line 2"]


class TestingTestcase(Bench):
    """
    Testcase class for testing all exception cases
    """
    def __init__(self, testStepFail=False, testStepError=False,
                 nameError=False, valueError=False, kbinterrupt=False,
                 exception=False, inconclusive_error=False, inRampUp=False,
                 inCase=False, inRampDown=False, test_step_timeout=False):
        self.testStepError = testStepError
        self.testStepFail = testStepFail
        self.nameError = nameError
        self.valueError = valueError
        self.kbinterrupt = kbinterrupt
        self.exception = exception
        self.inRampUp = inRampUp
        self.inCase = inCase
        self.inRampDown = inRampDown
        self.inconclusive = inconclusive_error
        self.test_step_timeout = test_step_timeout
        Bench.__init__(self,
                       name="ut_exception",
                       title = "unittest exception in testcase",
                       status="development",
                       type="acceptance",
                       purpose = "dummy",
                       requirements={
                           "duts": {
                               '*': { #requirements for all nodes
                                    "count":0,
                                }
                           }}
        )


    def raiseExc(self):
        if self.testStepFail:
            raise TestStepFail("This is a TestStepFail")
        if self.testStepError:
            raise TestStepError("This is a TestStepError")
        elif self.nameError:
            raise NameError("This is a NameError")
        elif self.valueError:
            raise ValueError("This is a ValueError")
        elif self.exception:
            raise Exception( "This is a generic exception" )
        elif self.kbinterrupt:
            raise KeyboardInterrupt()
        elif self.inconclusive:
            raise InconclusiveError("This will result in an inconclusive retcode.")
        elif self.test_step_timeout:
            raise TestStepTimeout("This is TestStepTimeout")

    def rampUp(self):
        self.args.silent = True
        if self.inRampUp:
            self.raiseExc()

    def case(self):
        if self.inCase:
            self.raiseExc()

    def rampDown(self):
        if self.inRampDown:
            self.raiseExc()

class TestVerify(unittest.TestCase):
    def setUp(self):
        logging.disable(logging.FATAL)

    def test_exceptions_in_case(self):
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(exception=True, inCase=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(testStepFail=True, inCase=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(testStepError=True, inCase=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(nameError=True, inCase=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(valueError=True, inCase=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(test_step_timeout=True, inCase=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        result = TestingTestcase(kbinterrupt=True, inCase=True).run()
        self.assertNotEqual( result, 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        self.assertTrue(result in ReturnCodes.INCONCLUSIVE_RETCODES, "Test return code not in inconclusive!")
        result = TestingTestcase(inconclusive_error=True, inCase=True).run()
        self.assertNotEqual(result, 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        self.assertTrue(result in ReturnCodes.INCONCLUSIVE_RETCODES, "Test return code not in inconclusive!")

    def test_verifyTrace(self):
        b = Bench()
        b.duts.append(MockDut())
        b.verify_trace(1, "this is test line 2")
        b.verify_trace(1, ["this is test line 2"])
        with self.assertRaises(LookupError) as e:
            b.verify_trace(1, "This is not found in traces")

    def test_exceptions_in_rampup(self):
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(exception=True, inRampUp=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(testStepFail=True, inRampUp=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(testStepError=True, inRampUp=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(nameError=True, inRampUp=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(valueError=True, inRampUp=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(kbinterrupt=True, inRampUp=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        result = TestingTestcase(inconclusive_error=True, inRampUp=True).run()
        self.assertNotEqual(result, 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        self.assertTrue(result in ReturnCodes.INCONCLUSIVE_RETCODES, "Test return code not in inconclusive!")

    @mock.patch("test.tests.test_tcTearDown.Testcase.teardown")
    def test_rampDown_called(self, mock_tearDown):
        retCode = TearDownTest(testStepFail=True).run()
        self.assertEquals(retCode, 1001)
        self.assertTrue(mock_tearDown.called)
        mock_tearDown.reset_mock()

        retCode = TearDownTest(testStepError=True).run()
        self.assertEquals(retCode, 1001)
        self.assertFalse(mock_tearDown.called)

        retCode = TearDownTest(exception=True).run()
        self.assertEquals(retCode, 1001)
        self.assertFalse(mock_tearDown.called)

        retCode = TearDownTest(testStepTimeout=True).run()
        self.assertEquals(retCode, 1001)
        self.assertFalse(mock_tearDown.called)

        # Test tearDown is called when TestStepTimeout raised in test case
        retCode = TearDownTest(testStepTimeoutInCase=True).run()
        self.assertEquals(retCode, 1005)
        self.assertTrue(mock_tearDown.called)
        mock_tearDown.reset_mock()

        retCode = TearDownTest(nameError=True).run()
        self.assertEquals(retCode, 1001)
        self.assertFalse(mock_tearDown.called)

        retCode = TearDownTest(valueError=True).run()
        self.assertEquals(retCode, 1001)
        self.assertFalse(mock_tearDown.called)

    def test_exceptions_in_rampdown(self):
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(exception=True, inRampDown=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(testStepFail=True, inRampDown=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(testStepError=True, inRampDown=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(nameError=True, inRampDown=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(valueError=True, inRampDown=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        n = threading.active_count()
        self.assertNotEqual( TestingTestcase(kbinterrupt=True, inRampDown=True).run(), 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        result = TestingTestcase(inconclusive_error=True, inRampDown=True).run()
        self.assertNotEqual(result, 0, "Test execution returned success retcode")
        self.assertEqual(n, threading.active_count())
        self.assertTrue(result in ReturnCodes.INCONCLUSIVE_RETCODES, "Test return code not in inconclusive!")

    def test_no_hanging_threads(self):
        n = threading.active_count()
        TestingTestcase().run()
        self.assertEqual(n, threading.active_count())

    @mock.patch('icetea_lib.bench.resource_provider', create=True)
    @mock.patch('icetea_lib.bench.execute_command', create=True)
    def test_precmds_to_two_duts(self, mock_ec, mock_rp):
        tc = Bench()
        tc._resource_provider = mock.Mock()
        tc.execute_command = mock.MagicMock()
        mock_resconf = mock.Mock()
        mock_resconf.get_dut_configuration = mock.MagicMock(return_value=[{"pre_cmds": ["first", "second"]}, {"pre_cmds": ["first2", "second2"]}])
        tc.resource_configuration = mock_resconf
        # Call using mangled name of __send_pre_commands method
        tc._Bench__send_pre_commands()
        tc.execute_command.assert_has_calls([mock.call(1, "first"), mock.call(1, "second"), mock.call(2, "first2"), mock.call(2, "second2")])

        tc.execute_command.reset_mock()
        # Test again with argument cmds
        tc._Bench__send_pre_commands("somecommand")
        tc.execute_command.assert_has_calls([mock.call(1, "first"), mock.call(1, "second"),
                                            mock.call(2, "first2"), mock.call(2, "second2"),
                                            mock.call("*", "somecommand")])

    @mock.patch('icetea_lib.bench.resource_provider', create=True)
    @mock.patch('icetea_lib.bench.execute_command', create=True)
    def test_postcmds_to_two_duts(self, mock_ec, mock_rp):
        tc = Bench()
        tc._resource_provider = mock.Mock()
        tc.execute_command = mock.MagicMock()
        mock_resconf = mock.Mock()
        mock_resconf.get_dut_configuration = mock.MagicMock(
            return_value=[{"post_cmds": ["first", "second"]}, {"post_cmds": ["first2", "second2"]}])
        tc.resource_configuration = mock_resconf
        # Call using mangled name of __send_pre_commands method
        tc._Bench__send_post_commands()
        tc.execute_command.assert_has_calls(
            [mock.call(1, "first"), mock.call(1, "second"), mock.call(2, "first2"), mock.call(2, "second2")])

        tc.execute_command.reset_mock()
        # Test again with argument cmds
        tc._Bench__send_post_commands("somecommand")
        tc.execute_command.assert_has_calls([mock.call(1, "first"), mock.call(1, "second"),
                                            mock.call(2, "first2"), mock.call(2, "second2"),
                                            mock.call("*", "somecommand")])

    def test_reset_duts(self):
        bench = Bench()
        mock_dut = mock.MagicMock()
        dutconf = {"reset.return_value": True, "initCLI.return_value": True}
        mock_dut.configure_mock(**dutconf)
        mock_dutrange = range(2)
        mock_duts = [mock_dut, mock_dut]
        bench.duts = mock_duts
        with mock.patch.object(bench, "get_dut_range", return_value=mock_dutrange):
            with mock.patch.object(bench, "is_my_dut", return_value=True):
                for method in ["hard", "soft", None]:
                    bench.args.reset = method
                    bench.reset_dut()
                    mock_dut.reset.assert_called_with(method)
                    self.assertEquals(mock_dut.reset.call_count, 2)
                    mock_dut.reset.reset_mock()

    def test_check_skip(self):
        tc = TestingTestcase()
        tc.config["requirements"]["duts"]["*"]["type"] = "process"
        tc.config["execution"] = {"skip": {"value": True, "only_type": "process"}}
        self.assertEqual(tc.run(), -1)
        self.assertTrue(tc.get_result().skipped())

    def test_check_skip_invalid_platform(self):
        tc = TestingTestcase()
        tc.config["requirements"]["duts"]["*"]["allowed_platforms"] = ["K64F"]
        tc.config["execution"] = {"skip": {"value": True, "platforms": ["K64F", "K65F"]}}
        self.assertEqual(tc.run(), 0)
        self.assertFalse(tc.get_result().skipped())

    def test_check_skip_valid_platform(self):
        tc = TestingTestcase()
        tc.config["requirements"]["duts"]["*"]["allowed_platforms"] = ["K64F"]
        tc.config["requirements"]["duts"]["*"]["platform_name"] = "K64F"
        tc.config["execution"] = {"skip": {"value": True, "platforms": ["K64F"]}}
        self.assertEqual(tc.run(), -1)
        self.assertTrue(tc.get_result().skipped())

    def test_check_no_skip(self):
        tc = TestingTestcase()
        tc.config["execution"] = {"skip": {"value": True}}
        self.assertEqual(tc.run(), 0)
        self.assertFalse(tc.get_result().skipped())

    def test_config_validation_bin_not_defined(self):
        duts_cfg = [{}]
        logger = logging.getLogger('unittest')
        logger.addHandler(logging.NullHandler())
        self.assertEqual(Bench._validate_dut_configs(duts_cfg, logger), None)

    def test_config_validation_bin_defined_but_not_exists(self):
        duts_cfg = [{"application": {"bin": "not.exist"}}]
        logger = logging.getLogger('unittest')
        logger.addHandler(logging.NullHandler())
        with self.assertRaises(EnvironmentError) as e:
            Bench._validate_dut_configs(duts_cfg, logger)

    def test_config_validation_bin_defined(self):
        duts_cfg = [{"application": {"bin": "./test/test_bench.py"}}]
        logger = logging.getLogger('unittest')
        logger.addHandler(logging.NullHandler())
        self.assertEqual(Bench._validate_dut_configs(duts_cfg, logger), None)

    def test_create_new_result(self):
        test_data = dict()
        test_data["reason"] = "this is a reason"
        result = Bench.create_new_result("fail", 1, 10, test_data)
        self.assertTrue(result.failure)

    def test_create_and_add_new_result(self):
        test_data = dict()
        test_data["reason"] = "this is a reason"
        bench = Bench()
        bench.add_new_result("fail", 1, 10, test_data)
        self.assertEqual(len(bench._results), 1)


if __name__=='__main__':
    unittest.main()



# pylint: disable=missing-docstring

"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest
from icetea_lib.CliResponse import CliResponse


class TestVerify(unittest.TestCase):

    def test_resps(self):
        response = CliResponse()
        response.lines = [
            'aapeli',
            'beeveli',
            'oopeli',
            'huhheli']
        self.assertTrue(response.verify_message(['oopeli'], False))
        self.assertTrue(response.verify_message(['oopeli', 'huhheli'], False))
        self.assertFalse(response.verify_message(['huhheli', 'oopeli'], False))

        with self.assertRaises(LookupError):
            self.assertTrue(response.verify_message(['oop eli']))

    def test_traces(self):
        response = CliResponse()
        response.traces = [
            'aapeli',
            'beeveli',
            'oopeli',
            'huhheli']
        self.assertTrue(response.verify_trace(['oopeli'], False))
        self.assertTrue(response.verify_trace(['oopeli', 'huhheli'], False))
        self.assertFalse(response.verify_trace(['huhheli', 'oopeli'], False))

        with self.assertRaises(LookupError):
            self.assertTrue(response.verify_trace(['oop eli']))

    def test_traces_deprecated(self):
        response = CliResponse()
        response.traces = [
            'aapeli',
            'beeveli',
            'oopeli',
            'huhheli']
        self.assertTrue(response.verify_trace(['oopeli'], False))
        self.assertTrue(response.verify_trace(['oopeli', 'huhheli'], False))
        self.assertFalse(response.verify_trace(['huhheli', 'oopeli'], False))

        with self.assertRaises(LookupError):
            self.assertTrue(response.verify_trace(['oop eli']))



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
import os

from icetea_lib.tools.file import FileUtils


def initLogger(name):
    '''
    Initializes a basic logger. Can be replaced when constructing the
    HttpApi object or afterwards with setter
    '''
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    #Skip attaching StreamHandler if one is already attached to logger
    if not getattr(logger, "streamhandler_set", None):
        ch = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        ch.setLevel(logging.INFO)
        logger.addHandler(ch)
        logger.streamhandler_set = True
    return logger


class JsonFile(object):
    def __init__(self, logger = None, filepath=None, filename=None):
        self.logger = logger if logger else initLogger("JsonFile")
        self.filepath = filepath if filepath else os.path.sep
        self.filename = filename if filename else "default_file.json"

    def write_file(self, content, filepath=None, filename=None, indent=None, keys_to_write=None):
        '''
        Write a Python dictionary as JSON to a file.

        :param content: Dictionary of key-value pairs to save to a file
        :param filepath: Path where the file is to be created
        :param filename: Name of the file to be created
        :param indent: You can use this to specify indent level for pretty printing the file
        :param keys_to_write: array of keys that are to be picked from data and written to file.
                Default is None, when all data is written to file.
        :return: Path of file used
        :raises OSError, EnvironmentError, ValueError
        '''
        path = filepath if filepath else self.filepath
        name = filename if filename else self.filename
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                self.logger.error("Error while creating directory: {}".format(e))
                raise

        name = self._ends_with(name, ".json")
        path = self._ends_with(path, os.path.sep)

        if keys_to_write:
            data_to_write = {}
            for key in keys_to_write:
                data_to_write[key] = content[key]
        else:
            data_to_write = content

        try:
            self._write_json(path, name, 'w', data_to_write, 2)
            return os.path.join(path, name)
        except EnvironmentError as e:
            self.logger.error("Error while opening or writing to file: {}".format(e))
            raise
        except ValueError as e:
            raise

    def read_file(self, filepath=None, filename=None):
        """
        Tries to read JSON content from filename and convert it to a dict.

        :param filepath: Path where the file is
        :param filename: File name
        :return: Dictionary read from the file
        :raises EnvironmentError, ValueError
        """
        name = filename if filename else self.filename
        path = filepath if filepath else self.filepath
        name = self._ends_with(name, ".json")
        path = self._ends_with(path, os.path.sep)

        try:
            return self._read_json(path, name)
        except EnvironmentError as e:
            self.logger.error("Error while opening or reading the file: {}".format(e))
            raise
        except ValueError as e:
            self.logger.error("File contents cannot be decoded to JSON: {}".format(e))
            raise

    def read_value(self, key, filepath=None, filename=None):
        """
        Tries to read the value of given key from JSON file filename.

        :param filepath: Path to file
        :param filename: Name of file
        :param key: Key to search for
        :return: Value corresponding to given key
        :raises OSError, EnvironmentError, KeyError
        """
        path = filepath if filepath else self.filepath
        name = filename if filename else self.filename
        name = self._ends_with(name, ".json")
        path = self._ends_with(path, os.path.sep)

        try:
            output = self._read_json(path, name)
            if key not in output:
                raise KeyError("Key '{}' not found in file {}".format(key, filename))
            else:
                return output[key]
        except EnvironmentError as e:
            self.logger.error("Error while opening or reading the file: {}".format(e))
            raise

    def write_values(self, data, filepath=None, filename=None, indent=None, keys_to_write=None):
        """
        Tries to write extra content to a JSON file.

        Creates filename.temp with updated content, removes the old file and
        finally renames the .temp to match the old file.
        This is in effort to preserve the data in case of some weird errors cause problems.

        :param filepath: Path to file
        :param filename: Name of file
        :param data: Data to write as a dictionary
        :param indent: indent level for pretty printing the resulting file
        :param keys_to_write: array of keys that are to be picked from data and written to file.
        Default is None, when all data is written to file.
        :return: Path to file used
        :raises EnvironmentError ValueError
        """

        name = filename if filename else self.filename
        path = filepath if filepath else self.filepath
        name = self._ends_with(name, ".json")
        path = self._ends_with(path, os.path.sep)

        if not os.path.isfile(path + name):
            try:
                return self.write_file(data, path, name, indent, keys_to_write)
            except EnvironmentError as e:
                self.logger.error("Error while opening or writing to file: {}".format(e))
                raise
            except ValueError as e:
                raise

        if keys_to_write:
            data_to_write = {}
            for key in keys_to_write:
                data_to_write[key] = data[key]
        else:
            data_to_write = data

        try:
            with open(path + name, 'r') as fil:
                output = json.load(fil)
                self.logger.info("Read contents of {}".format(filename))
                for key in data_to_write:
                    try:
                        output[key] = data_to_write[key]
                    except TypeError as e:
                        self.logger.error("File contents could not be serialized into a dict. {}".format(e))
                        raise
            self._write_json(path, name + ".temp", "w", output, indent)
            FileUtils.remove_file(name, path)
            FileUtils.rename_file(name + '.temp', name, path)
            return os.path.join(path, name)
        except EnvironmentError as e:
            self.logger.error("Error while writing to, opening or reading the file: {}".format(e))
            raise
        except ValueError as e:
            self.logger.error("File could not be decoded to JSON. It might be empty? {}".format(e))
            try:
                self._write_json(path, name, "w", data_to_write, indent)
                return os.path.join(path, name)
            except EnvironmentError:
                raise

    def _write_json(self, filepath, filename, writemode, content, indent):
        with open(os.path.join(filepath, filename), writemode) as fil:
            json.dump(content, fil, indent=indent)
            self.logger.info("Wrote content to file {}".format(filename))

    def _read_json(self, path, name):
        with open(os.path.join(path, name), 'r') as fil:
            output = json.load(fil)
            self.logger.info("Read contents of {}".format(name))
            return output

    def _ends_with(self, string, end):
        if not string.endswith(end):
            return string + end
        else:
            return string



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring

"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import os
import subprocess
import sys
import unittest

from test.dummy_dut import compile_dummy_dut
from icetea_lib.IceteaManager import IceteaManager


LIBPATH = os.path.abspath(os.path.join(os.path.dirname(__file__), ".." + os.path.sep))
sys.path.append(LIBPATH)


class TestFileTestCase(unittest.TestCase):
    def setUp(self):
        self.ctm = IceteaManager()

        # variables for testing getLocalTestcases, parseLocalTestcases,
        # parseLocalTest, loadClass, printListTestcases

        self.testpath = os.path.abspath(os.path.dirname(__file__))
        self.root_path = os.getcwd()
        sys.path.append(self.testpath)
        self.testdir = os.path.join(self.testpath, 'testbase')
        #variables for testing run()

        compile_dummy_dut()

    def test_init_with_non_existing_file(self):  # pylint: disable=invalid-name
        proc = subprocess.Popen(['python', 'icetea.py', '-s', '--tc', 'test_cmdline',
                                 '--tcdir', 'examples', '--type', 'process', '--clean',
                                 '--bin', 'file.bin'],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                cwd=self.root_path)
        out, error = proc.communicate()  # pylint: disable=unused-variable
        self.assertTrue(out.find(b"Binary not found") != -1,
                        "non exitent file error was not risen")
        self.assertEqual(proc.returncode, 0)

    @unittest.skipIf(sys.platform == 'win32', "windows doesn't support process test")
    def test_init_with_an_existing_file(self):
        bin_path = "test" + os.path.sep + "dut" + os.path.sep + "dummyDut"
        proc = subprocess.Popen(['python', 'icetea.py', '-s', '--tc', 'test_cmdline',
                                 '--tcdir', 'examples', '--type', 'process',
                                 '--clean', '--bin', bin_path],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                cwd=self.root_path)
        out, error = proc.communicate()  # pylint: disable=unused-variable
        self.assertTrue(out.find(b'test_cmdline |   pass') != -1, "exitent file was not accepted")
        self.assertEqual(proc.returncode, 0, "Icetea execution with existing file crashed")


if __name__ == '__main__':
    unittest.main()



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from os.path import abspath, relpath, dirname
import subprocess
import re

from icetea_lib.tools.tools import IS_PYTHON3


def get_path(filename):
    """
    Get absolute path for filename.

    :param filename: file
    :return: path
    """
    path = abspath(filename) if os.path.isdir(filename) else dirname(abspath(filename))
    return path


def get_git_root(path):
    """
    Run git rev-parse --show-toplevel in path.

    :param path: path where to run git
    :return: str
    """
    path = get_path(path)
    return __run_git(["rev-parse", "--show-toplevel"], path)[0]


def get_git_file_path(filename):
    """
    Get relative path for filename in git root.

    :param filename: File name
    :return: relative path or None
    """
    git_root = get_git_root(filename)
    return relpath(filename, git_root).replace("\\", "/") if git_root else ''


def get_commit_id(path):
    """
    Run git rev-parse HEAD in path.

    :param path: Path where git is to be run.
    :return: str or None
    """
    path = get_path(path)
    return __run_git(["rev-parse", "HEAD"], path)[0]


def get_remote_url(path, remote="origin"):
    """
    Run git config --get remote.<remote>.url in path.

    :param path: Path where git is to be run
    :param remote: Remote name
    :return: str or None
    """
    path = get_path(path)
    cmd = ["config", "--get", "remote.%s.url" % remote]
    return __run_git(cmd, path)[0]


def is_git_root_dirty(path):
    """
    Run git diff-index --quiet HEAD in path.

    :param path: path where git is to be run
    :return: str or None.
    """
    path = get_path(path)
    cmd = ["diff-index", "--quiet", "HEAD"]
    return __run_git(cmd, path)[1] != 0


def get_current_branch(path):
    """
    Run git rev-parse --abbrev-ref HEAD in path.

    :param path: path where git is to be run
    :return: str or None.
    """
    path = get_path(path)
    exe = ["rev-parse", "--abbrev-ref", "HEAD"]
    return __run_git(exe, path)[0]


def get_git_info(git_folder, verbose=False):
    """
    Detect GIT information by folder.

    :param git_folder: Folder
    :param verbose: Verbosity, boolean, default is False
    :return: dict
    """
    if verbose:
        print("detect GIT info by folder: '%s'" % git_folder)
    try:
        git_info = {
            "commitid": get_commit_id(git_folder),
            "branch": get_current_branch(git_folder),
            "git_path": get_git_file_path(git_folder),
            "url": get_remote_url(git_folder),
            "scm": "unknown",
            "scm_group": "unknown",
            "scm_path": "unknown",
            "scm_link": ""
        }
        if is_git_root_dirty(git_folder):
            git_info['dirty'] = True
    except Exception as err:  # pylint: disable=broad-except
        print("GitTool exception:")
        print(err)
        return {}

    if isinstance(git_info['url'], str):
        match = re.search(r"github\.com:(.*)\/(.*)", git_info['url'])
        if match:
            git_info["scm"] = "github.com"
            git_info["scm_path"] = match.group(2)
            git_info["scm_group"] = match.group(1)
            scm_link_end = " %s/%s" % (git_info["scm_group"],
                                       git_info["scm_path"].replace(".git", ""))
            git_info["scm_link"] = "https://github.com/" + scm_link_end
            git_info["scm_link"] += "/tree/%s/%s" % (git_info['commitid'], git_info["git_path"])

    if verbose:
        print("all git_info:")
        print(git_info)

    return git_info


def __get_git_bin():
    """
    Get git binary location.

    :return: Check git location
    """
    git = 'git'
    alternatives = [
        '/usr/bin/git'
    ]
    for alt in alternatives:
        if os.path.exists(alt):
            git = alt
            break
    return git


def __run_git(cmd, path=None):
    """internal run git command
    :param cmd: git parameters as array
    :param path: path where command will be executed
    :return: tuple (<line>, <returncode>)
    """
    exe = [__get_git_bin()] + cmd
    try:
        proc = subprocess.Popen(exe, cwd=path, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return None, None
    except ValueError:
        return None, None
    except OSError:
        return None, None

    out, err = proc.communicate()
    if IS_PYTHON3:
        out = out.decode("utf-8")
    if err:
        print("Cmd ('%s') fails: %s" % (' '.join(exe), err))
        return None, proc.returncode
    return out.strip(), proc.returncode


if __name__ == "__main__":
    import sys
    import json
    FILE_NAME = sys.argv[1] if sys.argv.count > 1 else "./../setup.py"
    print(json.dumps(get_git_info(FILE_NAME), sort_keys=True, indent=4))



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

ResultList module, contains the ResultList class, which is an object to store Result objects.
"""

from collections import Iterator

from icetea_lib.Reports.ReportJunit import ReportJunit
from icetea_lib.Reports.ReportHtml import ReportHtml
from icetea_lib.Reports.ReportConsole import ReportConsole
from icetea_lib.Result import Result


class ResultList(Iterator):
    """
    List of Result objects. Implements parts of Iterator interface to allow some ease of use.
    """
    def __init__(self):  # pylint: disable=super-init-not-called
        """
        Constructor for ResultList
        """
        self.data = []
        self.index = 0

    def append(self, result):
        """
        Append a new Result to the list
        :param result: Result to append
        :return: Nothing
        :raises: TypeError if result is not Result or ResultList
        """
        if isinstance(result, Result):
            self.data.append(result)
        elif isinstance(result, ResultList):
            self.data += result.data
        else:
            raise TypeError('unknown result type')

        # @todo this could be used to generate html table after each test..
        # self._save_html_report({"NOTE": "TESTS EXECUTION IS ONGOING.."}, reload=5)

    def save(self, heads):
        """
        Create reports in different formats.

        :param heads: html table extra values in title rows
        """
        # Junit
        self._save_junit()
        # HTML
        self._save_html_report(heads)
        # Console print
        self._print_console_summary()

    def _save_junit(self):
        """
        Save Junit report.

        :return: Nothing
        """
        report = ReportJunit(self)
        file_name = report.get_latest_filename("result.junit.xml", "")
        report.generate(file_name)

        file_name = report.get_latest_filename("junit.xml", "../")
        report.generate(file_name)

    def _save_html_report(self, heads=None, refresh=None):
        """
        Save html report.

        :param heads: headers as dict
        :param refresh: Boolean, if True will add a reload-tag to the report
        :return: Nothing
        """
        report = ReportHtml(self)
        heads = heads if heads else {}
        test_report_filename = report.get_current_filename("html")
        report.generate(test_report_filename, title='Test Results', heads=heads, refresh=refresh)

        # Update latest.html in the log root directory
        latest_report_filename = report.get_latest_filename("html")
        report.generate(latest_report_filename, title='Test Results', heads=heads, refresh=refresh)

    def _print_console_summary(self):
        """
        Print the console report.

        :return: Nothing
        """
        ReportConsole(self).generate()

    def success_count(self):
        """
        Amount of passed test cases in this list.

        :return: integer
        """
        return len([i for i, result in enumerate(self.data) if result.success])

    def failure_count(self):
        """
        Amount of failed test cases in this list.

        :return: integer
        """
        return len([i for i, result in enumerate(self.data) if result.failure])

    def inconclusive_count(self):
        """
        Amount of inconclusive test cases in this list.

        :return: integer
        """
        inconc_count = len([i for i, result in enumerate(self.data) if result.inconclusive])
        unknown_count = len([i for i, result in enumerate(self.data) if result.get_verdict() ==
                             "unknown"])
        return inconc_count + unknown_count

    def skip_count(self):
        """
        Amount of skipped test cases in this list.

        :return: integer
        """
        return len([i for i, result in enumerate(self.data) if result.skip])

    @property
    def skipped(self):
        return True if self.skip_count() == len(self) else False

    @property
    def inconclusive(self):
        return True if self.inconclusive_count() and not self.failure else False

    @property
    def success(self):
        return True if not self.inconclusive and not self.failure else False

    @property
    def failure(self):
        for item in self.data:
            if item.failure:
                return True
        return False

    def get_verdict(self):
        if self.success:
            return "pass"
        elif self.failure:
            return "fail"
        elif self.skipped:
            return "skip"
        else:
            return "inconclusive"

    def total_duration(self):
        """
        Sum of the durations of the tests in this list.

        :return: integer
        """
        durations = [result.duration for result in self.data]
        return sum(durations)

    def pass_rate(self, include_skips=False, include_inconclusive=False):
        """
        Calculate pass rate for tests in this list.

        :param include_skips: Boolean, if True skipped tc:s will be included. Default is False
        :param include_inconclusive: Boolean, if True inconclusive tc:s will be included.
        Default is False.
        :return: Percentage in format .2f %
        """
        total = self.count()
        success = self.success_count()
        try:
            if include_inconclusive and include_skips:
                val = 100.0*success/total
            elif include_inconclusive:
                inconcs = self.inconclusive_count()
                val = 100.0*success/(total-inconcs)
            elif include_skips:
                skipped = self.skip_count()
                val = 100.0*success/(total-skipped)
            else:
                failures = self.failure_count()
                val = 100.0*success/(failures+success)
        except ZeroDivisionError:
            val = 0
        return format(val, '.2f') + " %"

    def get_summary(self):
        """
        Get a summary of this ResultLists contents as dictionary.

        :return: dictionary
        """
        return {
            "count": self.count(),
            "pass": self.success_count(),
            "fail": self.failure_count(),
            "skip": self.skip_count(),
            "inconclusive": self.inconclusive_count(),
            "duration": self.total_duration(),
        }

    def count(self):
        """
        Return length of this list (amount of Results).

        :return: integer
        """
        return len(self)

    def __len__(self):
        """
        Return length of this list (amount of Results).

        :return: integer
        """
        return len(self.data)

    def __next__(self):
        """
        Get next Result from this list.

        :return: Result
        """
        return self.next()

    def next(self):
        """
        Implementation of next method from Iterator.

        :return: Result
        :raises: StopIteration if IndexError occurs.
        """
        try:
            result = self.data[self.index]
        except IndexError:
            self.index = 0
            raise StopIteration
        self.index += 1
        return result



"""
mbed SDK
Copyright (c) 2011-2015 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Authors:
Jussi Vatjus-Anttila <jussi.vatjus-anttila@arm.com>
"""

import re
import os
import requests
from icetea_lib.tools.tools import sha1_of_file


class NotFoundError(Exception):
    """
    Not Found Error
    """
    pass


class Build(object):
    """
    Class Build
    """
    def __init__(self, **kwargs):
        """
        :param kwargs: dictionary of ref, type
        :return: nothing
        """
        self._ref = kwargs['ref']
        self._type = kwargs['type']
        self.name = None
        self.date = None
        self.sha1 = None
        self.branch = None
        self.toolchain = None
        self.build_url = None
        self.giturl = None
        self.commit_id = None

    @staticmethod
    def init(ref):
        """function init
        ref: reference for build, e.g. file:c:\\.. or http://...
        """
        r_http = re.compile(r"^(https?:.*)")
        r_uuid = re.compile(
            r"^([a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12})$",
            re.I)
        r_id = re.compile(r"^([0-9a-fA-F]{24})$")
        r_file = re.compile(r"^file:(.*)$")
        r_any = re.compile(r"([\S\s]{4,})$")
        types = [
            {"cls": BuildHttp, "res": [r_http]},
            {"cls": BuildDatabase, "res": [r_uuid, r_id]},
            {"cls": BuildFile, "res": [r_file, r_any]}
        ]

        for _type in types:
            for reg in _type["res"]:
                match = reg.match(ref)
                if match:
                    return _type["cls"](ref=match.group(1))

        raise NotImplementedError("Cannot detect reference type: %s" % ref)

    def get_data(self):
        """
        function get_data
        :return: Buffer
        """
        return self._load()

    def get_type(self):
        """
        :return: reference type - file,id,url
        """
        return self._type

    def get_url(self):
        """
        Get reference url.

        :return: reference type - file,id,url
        """
        return self._ref

    def get_file(self):
        """
        Load data into a file and return file path.

        :return: path to file as string
        """
        raise NotImplementedError()


    def _load(self):
        """
        Load test case
        """
        raise NotImplementedError()

    @property
    def url(self):
        return self._ref

    @url.setter
    def url(self, url):
        self._ref = url


class BuildFile(Build):
    """
    Build as a File
    """

    def __init__(self, ref):
        """
        Constructor
        """
        super(BuildFile, self).__init__(ref=ref, type='file')
        sha = sha1_of_file(ref)
        self.sha1 = sha if sha else ""

    def is_exists(self):
        """
        Check if file exists.

        :return: Boolean
        """
        return os.path.isfile(self._ref)

    def get_file(self):
        return self._ref if self.is_exists() else None

    def _load(self):
        """
        Function load.

        :return: file contents
        :raises: NotFoundError if file not found
        """
        if self.is_exists():
            return open(self._ref, "rb").read()
        raise NotFoundError("File %s not found" % self._ref)

class BuildHttp(Build):
    """
    Build as a Http link
    """

    def __init__(self, ref):
        super(BuildHttp, self).__init__(ref=ref, type='http')
        self.auth = ('user', 'pass')
        self.http_verify = False
        self.timeout = None

    def _load(self):
        """
        Function load.
        :return: Response content
        :raises: NotFoundError
        """
        try:
            get = requests.get(self._ref,
                               verify=self.http_verify,
                               auth=self.auth,
                               timeout=self.timeout)
        except requests.exceptions.RequestException as err:
            raise NotFoundError(err)
        return get.content

    def is_exists(self):
        """
        Check if file exists
        :return: Boolean
        """
        try:
            return self._load() is not None
        except NotFoundError:
            return False

    def get_file(self):
        """
        Load data into a file and return file path.

        :return: path to file as string
        """
        content = self._load()
        if not content:
            return None
        filename = "temporary_file.bin"
        with open(filename, "wb") as file_name:
           file_name.write(content)
        return filename


class BuildDatabase(Build):
    """
    Build as a Database ID
    """

    def __init__(self, ref):
        super(BuildDatabase, self).__init__(ref=ref, type='database')

    def _load(self):
        """function load
        returns
        """
        raise NotImplementedError("loading from DB not supported")



"""
Copyright 2017 ARM Limited
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

DutInformation module. Contains DutInformation and DutInformationList classes.
These classes are meant to contain information of allocated duts. DutInformationList contains
several helper methods to get some relevant information from it's contents.
"""


#Disable too many arguments warning, too few public methods warning
#pylint: disable=R0913,R0903

from jsonmerge import merge

from icetea_lib.ResourceProvider.exceptions import ResourceInitError


class DutInformation(object):
    """
    DutInformation object.
    Contains fields platform (string), index (int), vendor (string), and build (Build)
    """
    def __init__(self, platform, resourceid, index="", vendor="", build=None, provider=None):
        self.platform = platform
        self.index = index
        self.resource_id = resourceid
        self.vendor = vendor
        self.build = build
        self.provider = provider
        if resourceid:
            DutInformationList.push_resource_cache(resourceid, self.as_dict())

    def as_dict(self):
        """
        Generate a dictionary of the contents of this DutInformation object.

        :return: dict
        """
        my_info = {}
        if self.platform:
            my_info["model"] = self.platform
        if self.resource_id:
            my_info["sn"] = self.resource_id
        if self.vendor:
            my_info["vendor"] = self.vendor
        if self.provider:
            my_info["provider"] = self.provider
        return my_info

    @property
    def build_binary_sha1(self):
        """
        Get flashed binary sha1 or None if not flashed yet.

        :return: dict object
        """
        cache = DutInformationList.get_resource_cache(self.resource_id)
        return cache.get("build_binary_sha1")

    @build_binary_sha1.setter
    def build_binary_sha1(self, value):
        """
        Setter for flashed binary sha1.

        :return: Nothing
        """
        DutInformationList.push_resource_cache(self.resource_id, {"build_binary_sha1": value})


class DutInformationList(object):
    """
    DutInformationList object. List of DutInformation objects in member dutinformations as a list.
    helper methods for getting unique dut models in either string or list format and a list of
    resource ID:s.
    """

    _cache = dict()

    def __init__(self, content=None):
        self.dutinformations = content if content else []

    def get(self, index):
        #Disable len(seq) as condition value warning
        #pylint: disable=C1801
        if index > len(self):
            return None
        return self.dutinformations[index]

    def get_uniq_string_dutmodels(self):
        """
        Gets a string of dut models in this TC.

        :return: String of dut models separated with commas.
        String "unknown platform" if no dut information is available
        """
        models = self.get_uniq_list_dutmodels()
        if not models:
            return ""
        return ",".join(models)

    def get_uniq_list_dutmodels(self):
        """
        Gets a list of dut models in this TC
        :return: List of dut models in this TC. Empty list if information is not available.
        """
        models = []
        if self.dutinformations:
            for info in self.dutinformations:
                models.append(info.platform)
            seen = []
            for item in models:
                if item not in seen:
                    seen.append(item)
            return seen
        return models

    def get_resource_ids(self):
        """
        Get resource ids as a list.

        :return: List of resource id:s or "unknown"
        """
        resids = []
        if self.dutinformations:
            for info in self.dutinformations:
                resids.append(info.resource_id)
            return resids
        return "unknown"

    def append(self, dutinfo):
        """
        Append a DutInformation object to the list.

        :param dutinfo: object to append
        :return: Nothing
        """
        self.dutinformations.append(dutinfo)

    def __len__(self):
        """
        overrides len operation for DutInformationList.

        :return: Length of internal dutinformation list as int
        """
        return len(self.dutinformations)

    @staticmethod
    def push_resource_cache(resourceid, info):
        """
        Cache resource specific information

        :param resourceid: Resource id as string
        :param info: Dict to push
        :return: Nothing
        """
        if not resourceid:
            raise ResourceInitError("Resource id missing")
        if not DutInformationList._cache.get(resourceid):
            DutInformationList._cache[resourceid] = dict()
        DutInformationList._cache[resourceid] = merge(DutInformationList._cache[resourceid], info)

    @staticmethod
    def get_resource_cache(resourceid):
        """
        Get a cached dictionary related to an individual resourceid.

        :param resourceid: String resource id.
        :return: dict
        """
        if not resourceid:
            raise ResourceInitError("Resource id missing")
        if not DutInformationList._cache.get(resourceid):
            DutInformationList._cache[resourceid] = dict()
        return DutInformationList._cache[resourceid]
